# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Lesser General Public License for more details.                   *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
from FreeCAD import Units
import Path
import argparse
import datetime
import shlex
import Path.Base.Util as PathUtil
import Path.Post.Utils as PostUtils
import PathScripts.PathUtils as PathUtils
from builtins import open as pyopen

TOOLTIP = """
This is a postprocessor file for the Path workbench. It is used to
take a pseudo-G-code fragment outputted by a Path object, and output
real G-code suitable for a linuxcnc 3 axis mill. This postprocessor, once placed
in the appropriate PathScripts folder, can be used directly from inside
FreeCAD, via the GUI importer or via python scripts with:

import linuxcnc_post
linuxcnc_post.export(object,"/path/to/file.ncc","")
"""

now = datetime.datetime.now()

parser = argparse.ArgumentParser(prog="linuxcnc", add_help=False)
parser.add_argument("--no-header", action="store_true", help="suppress header output")
parser.add_argument("--no-comments", action="store_true", help="suppress comment output")
parser.add_argument("--line-numbers", action="store_true", help="prefix with line numbers")
parser.add_argument(
    "--no-show-editor",
    action="store_true",
    help="don't pop up editor before writing output",
)
parser.add_argument("--precision", default="3", help="number of digits of precision, default=3")
parser.add_argument(
    "--preamble",
    help='set commands to be issued before the first command, default="G17 G54 G40 G49 G80 G90\\n"',
)
parser.add_argument(
    "--postamble",
    help='set commands to be issued after the last command, default="M05\\nG17 G54 G90 G80 G40\\nM2\\n"',
)
parser.add_argument(
    "--inches", action="store_true", help="Convert output for US imperial mode (G20)"
)
parser.add_argument(
    "--modal",
    action="store_true",
    help="Output the Same G-command Name USE NonModal Mode",
)
parser.add_argument("--axis-modal", action="store_true", help="Output the Same Axis Value Mode")
parser.add_argument(
    "--no-tlo",
    action="store_true",
    help="suppress tool length offset (G43) following tool changes",
)

TOOLTIP_ARGS = parser.format_help()

# These globals set common customization preferences
OUTPUT_COMMENTS = True
OUTPUT_HEADER = True
OUTPUT_LINE_NUMBERS = False
SHOW_EDITOR = True
MODAL = False  # if true commands are suppressed if the same as previous line.
USE_TLO = True  # if true G43 will be output following tool changes
OUTPUT_DOUBLES = True  # if false duplicate axis values are suppressed if the same as previous line.
COMMAND_SPACE = " "
LINENR = 100  # line number starting value

# These globals will be reflected in the Machine configuration of the project
UNITS = "G21"  # G21 for metric, G20 for us standard
UNIT_SPEED_FORMAT = "mm/min"
UNIT_FORMAT = "mm"

MACHINE_NAME = "LinuxCNC"
CORNER_MIN = {"x": 0, "y": 0, "z": 0}
CORNER_MAX = {"x": 500, "y": 300, "z": 300}
PRECISION = 3

# Preamble text will appear at the beginning of the GCODE output file.
PREAMBLE = """G17 G54 G40 G49 G80 G90
"""

# Postamble text will appear following the last operation.
POSTAMBLE = """M05
G17 G54 G90 G80 G40
M2
"""

# Pre operation text will be inserted before every operation
PRE_OPERATION = """"""

# Post operation text will be inserted after every operation
POST_OPERATION = """"""

# Tool Change commands will be inserted before a tool change
TOOL_CHANGE = """"""


def processArguments(argstring):
    global OUTPUT_HEADER
    global OUTPUT_COMMENTS
    global OUTPUT_LINE_NUMBERS
    global SHOW_EDITOR
    global PRECISION
    global PREAMBLE
    global POSTAMBLE
    global UNITS
    global UNIT_SPEED_FORMAT
    global UNIT_FORMAT
    global MODAL
    global USE_TLO
    global OUTPUT_DOUBLES

    try:
        args = parser.parse_args(shlex.split(argstring))
        if args.no_header:
            OUTPUT_HEADER = False
        if args.no_comments:
            OUTPUT_COMMENTS = False
        if args.line_numbers:
            OUTPUT_LINE_NUMBERS = True
        if args.no_show_editor:
            SHOW_EDITOR = False
        print("Show editor = %d" % SHOW_EDITOR)
        PRECISION = args.precision
        if args.preamble is not None:
            PREAMBLE = args.preamble.replace("\\n", "\n")
        if args.postamble is not None:
            POSTAMBLE = args.postamble.replace("\\n", "\n")
        if args.inches:
            UNITS = "G20"
            UNIT_SPEED_FORMAT = "in/min"
            UNIT_FORMAT = "in"
            PRECISION = 4
        if args.modal:
            MODAL = True
        if args.no_tlo:
            USE_TLO = False
        if args.axis_modal:
            print("here")
            OUTPUT_DOUBLES = False

    except Exception:
        return False

    return True


def export(objectslist, filename, argstring):
    if not processArguments(argstring):
        return None
    global UNITS
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT

    for obj in objectslist:
        if not hasattr(obj, "Path"):
            print(
                "the object " + obj.Name + " is not a path. Please select only path and Compounds."
            )
            return None

    print("postprocessing...")
    gcode = ""

    # write header
    if OUTPUT_HEADER:
        gcode += linenumber() + "(Exported by FreeCAD)\n"
        gcode += linenumber() + "(Post Processor: " + __name__ + ")\n"
        gcode += linenumber() + "(Output Time:" + str(now) + ")\n"

    # Write the preamble
    if OUTPUT_COMMENTS:
        gcode += linenumber() + "(begin preamble)\n"
    for line in PREAMBLE.splitlines():
        gcode += linenumber() + line + "\n"
    gcode += linenumber() + UNITS + "\n"

    for obj in objectslist:
        # Skip inactive operations
        if not PathUtil.activeForOp(obj):
            continue

        # do the pre_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(begin operation: %s)\n" % obj.Label
            gcode += linenumber() + "(machine units: %s)\n" % (UNIT_SPEED_FORMAT)
        for line in PRE_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # get coolant mode
        coolantMode = PathUtil.coolantModeForOp(obj)

        # turn coolant on if required
        if OUTPUT_COMMENTS:
            if not coolantMode == "None":
                gcode += linenumber() + "(Coolant On:" + coolantMode + ")\n"
        if coolantMode == "Flood":
            gcode += linenumber() + "M8" + "\n"
        if coolantMode == "Mist":
            gcode += linenumber() + "M7" + "\n"

        # process the operation gcode
        gcode += parse(obj)

        # do the post_op
        if OUTPUT_COMMENTS:
            gcode += linenumber() + "(finish operation: %s)\n" % obj.Label
        for line in POST_OPERATION.splitlines(True):
            gcode += linenumber() + line

        # turn coolant off if required
        if not coolantMode == "None":
            if OUTPUT_COMMENTS:
                gcode += linenumber() + "(Coolant Off:" + coolantMode + ")\n"
            gcode += linenumber() + "M9" + "\n"

    # do the post_amble
    if OUTPUT_COMMENTS:
        gcode += "(begin postamble)\n"
    for line in POSTAMBLE.splitlines():
        gcode += linenumber() + line + "\n"

    if FreeCAD.GuiUp and SHOW_EDITOR:
        final = gcode
        if len(gcode) > 100000:
            print("Skipping editor since output is greater than 100kb")
        else:
            dia = PostUtils.GCodeEditorDialog()
            dia.editor.setText(gcode)
            result = dia.exec_()
            if result:
                final = dia.editor.toPlainText()
    else:
        final = gcode

    print("done postprocessing.")

    if not filename == "-":
        gfile = pyopen(filename, "w")
        gfile.write(final)
        gfile.close()

    return final


def linenumber():
    global LINENR
    if OUTPUT_LINE_NUMBERS is True:
        LINENR += 10
        return "N" + str(LINENR) + " "
    return ""


def parse(pathobj):
    global PRECISION
    global MODAL
    global OUTPUT_DOUBLES
    global UNIT_FORMAT
    global UNIT_SPEED_FORMAT

    out = ""
    lastcommand = None
    precision_string = "." + str(PRECISION) + "f"
    currLocation = {}  # keep track for no doubles

    # the order of parameters
    # linuxcnc doesn't want K properties on XY plane  Arcs need work.
    params = [
        "X",
        "Y",
        "Z",
        "A",
        "B",
        "C",
        "I",
        "J",
        "F",
        "S",
        "T",
        "Q",
        "R",
        "L",
        "H",
        "D",
        "P",
    ]
    firstmove = Path.Command("G0", {"X": -1, "Y": -1, "Z": -1, "F": 0.0})
    currLocation.update(firstmove.Parameters)  # set First location Parameters

    if hasattr(pathobj, "Group"):  # We have a compound or project.
        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(compound: " + pathobj.Label + ")\n"
        for p in pathobj.Group:
            out += parse(p)
        return out
    else:  # parsing simple path
        # groups might contain non-path things like stock.
        if not hasattr(pathobj, "Path"):
            return out

        # if OUTPUT_COMMENTS:
        #     out += linenumber() + "(" + pathobj.Label + ")\n"

        # The following "for" statement was fairly recently added
        # but seems to be using the A, B, and C parameters in ways
        # that don't appear to be compatible with how the PATH code
        # uses the A, B, and C parameters.  I have reverted the
        # change here until we can figure out what it going on.
        #
        # for c in PathUtils.getPathWithPlacement(pathobj).Commands:
        for c in pathobj.Path.Commands:
            outstring = []
            command = c.Name
            outstring.append(command)

            # if modal: suppress the command if it is the same as the last one
            if MODAL is True:
                if command == lastcommand:
                    outstring.pop(0)

            if c.Name.startswith("(") and not OUTPUT_COMMENTS:  # command is a comment
                continue

            # Handle G84/G74 tapping cycles
            if command in ("G84", "G74") and "F" in c.Parameters:
                pitch_mm = float(c.Parameters["F"])
                c.Parameters.pop("F")  # Remove F from output, we'll handle it

                # Get spindle speed (from S param or last known value)
                spindle_speed = None
                if "S" in c.Parameters:
                    spindle_speed = float(c.Parameters["S"])
                    c.Parameters.pop("S")

                # Convert pitch to inches if needed
                if UNITS == "G20":  # imperial
                    pitch = pitch_mm / 25.4
                else:
                    pitch = pitch_mm

                # Calculate feed rate
                if spindle_speed is not None:
                    feed_rate = pitch * spindle_speed
                    speed = Units.Quantity(feed_rate, UNIT_SPEED_FORMAT)
                    outstring.append(
                        "F" + format(float(speed.getValueAs(UNIT_SPEED_FORMAT)), precision_string)
                    )
                else:
                    # No spindle speed found, output pitch as F
                    outstring.append("F" + format(pitch, precision_string))

            # Now add the remaining parameters in order
            for param in params:
                if param in c.Parameters:
                    if param == "F" and (
                        currLocation[param] != c.Parameters[param] or OUTPUT_DOUBLES
                    ):
                        if c.Name not in [
                            "G0",
                            "G00",
                        ]:  # linuxcnc doesn't use rapid speeds
                            speed = Units.Quantity(c.Parameters["F"], FreeCAD.Units.Velocity)
                            if speed.getValueAs(UNIT_SPEED_FORMAT) > 0.0:
                                outstring.append(
                                    param
                                    + format(
                                        float(speed.getValueAs(UNIT_SPEED_FORMAT)),
                                        precision_string,
                                    )
                                )
                        else:
                            continue
                    elif param == "T":
                        outstring.append(param + str(int(c.Parameters["T"])))
                    elif param == "H":
                        outstring.append(param + str(int(c.Parameters["H"])))
                    elif param == "D":
                        outstring.append(param + str(int(c.Parameters["D"])))
                    elif param == "S":
                        outstring.append(param + str(int(c.Parameters["S"])))
                    else:
                        if (
                            (not OUTPUT_DOUBLES)
                            and (param in currLocation)
                            and (currLocation[param] == c.Parameters[param])
                        ):
                            continue
                        else:
                            if param in ("A", "B", "C"):
                                outstring.append(
                                    param + format(float(c.Parameters[param]), precision_string)
                                )
                            else:
                                pos = Units.Quantity(c.Parameters[param], FreeCAD.Units.Length)
                                outstring.append(
                                    param
                                    + format(float(pos.getValueAs(UNIT_FORMAT)), precision_string)
                                )

            # store the latest command
            lastcommand = command
            currLocation.update(c.Parameters)

            # Check for Tool Change:
            if command == "M6":
                # stop the spindle
                out += linenumber() + "M5\n"
                for line in TOOL_CHANGE.splitlines(True):
                    out += linenumber() + line

                # add height offset
                if USE_TLO:
                    tool_height = "\nG43 H" + str(int(c.Parameters["T"]))
                    outstring.append(tool_height)

            if command == "message":
                if OUTPUT_COMMENTS is False:
                    out = []
                else:
                    outstring.pop(0)  # remove the command

            # prepend a line number and append a newline
            if len(outstring) >= 1:
                if OUTPUT_LINE_NUMBERS:
                    outstring.insert(0, (linenumber()))

                # append the line to the final output
                for w in outstring:
                    out += w + COMMAND_SPACE
                out += "\n"

        return out


# print(__name__ + " gcode postprocessor loaded.")



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2013 Yorik van Havre <yorik@uncreated.net>              *
# *   Copyright (c) 2019 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Unit tests for the Draft Workbench, DWG import and export tests."""

## @package test_dwg
# \ingroup drafttests
# \brief Unit tests for the Draft Workbench, DWG import and export tests.

## \addtogroup drafttests
# @{

import os

import FreeCAD as App
import Draft
from drafttests import auxiliary as aux
from drafttests import test_base
from draftutils.messages import _msg


class DraftDWG(test_base.DraftTestCaseDoc):
    """Test reading and writing of DWG files with Draft."""

    def test_read_dwg(self):
        """Read a DWG file and import its elements as Draft objects."""
        operation = "importDWG.import"
        _msg("  Test '{}'".format(operation))
        _msg("  This test requires a DWG file to read.")

        file = "Mod/Draft/drafttest/test.dwg"
        in_file = os.path.join(App.getResourceDir(), file)
        _msg("  file={}".format(in_file))
        _msg("  exists={}".format(os.path.exists(in_file)))

        obj = aux.fake_function(in_file)
        self.assertTrue(obj, "'{}' failed".format(operation))

    def test_export_dwg(self):
        """Create some figures and export them to a DWG file."""
        operation = "importDWG.export"
        _msg("  Test '{}'".format(operation))

        file = "Mod/Draft/drafttest/out_test.dwg"
        out_file = os.path.join(App.getResourceDir(), file)
        _msg("  file={}".format(out_file))
        _msg("  exists={}".format(os.path.exists(out_file)))

        obj = aux.fake_function(out_file)
        self.assertTrue(obj, "'{}' failed".format(operation))


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2022 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

"""GUI functions for uninstalling an Addon or Macro."""

import addonmanager_freecad_interface as fci
from Widgets.addonmanager_utility_dialogs import MessageDialog

# Get whatever version of PySide we can
try:
    from PySide import QtCore, QtWidgets  # Use the FreeCAD wrapper
except ImportError:
    try:
        from PySide6 import QtCore, QtWidgets  # Outside FreeCAD, try Qt6 first
    except ImportError:
        from PySide2 import QtCore, QtWidgets  # Fall back to Qt5

from addonmanager_uninstaller import AddonUninstaller, MacroUninstaller
import addonmanager_utilities as utils

translate = fci.translate


class AddonUninstallerGUI(QtCore.QObject):
    """User interface for uninstalling an Addon: asks for confirmation, displays a progress dialog,
    displays completion and/or error dialogs, and emits the finished() signal when all work is
    complete."""

    finished = QtCore.Signal()

    def __init__(self, addon_to_remove):
        super().__init__()
        self.addon_to_remove = addon_to_remove
        if hasattr(self.addon_to_remove, "macro") and self.addon_to_remove.macro is not None:
            self.uninstaller = MacroUninstaller(self.addon_to_remove)
        else:
            self.uninstaller = AddonUninstaller(self.addon_to_remove)
        self.uninstaller.success.connect(self._succeeded)
        self.uninstaller.failure.connect(self._failed)
        self.worker_thread = QtCore.QThread()
        self.worker_thread.setObjectName("AddonUninstallerGUI worker thread")
        self.uninstaller.moveToThread(self.worker_thread)
        self.uninstaller.finished.connect(self.worker_thread.quit)
        self.worker_thread.started.connect(self.uninstaller.run)
        self.progress_dialog = None
        self.dialog_timer = QtCore.QTimer()
        self.dialog_timer.timeout.connect(self._show_progress_dialog)
        self.dialog_timer.setSingleShot(True)
        self.dialog_timer.setInterval(1000)  # Can override from external (e.g. testing) code

    def run(self):
        """Begin the user interaction: asynchronous, only blocks while showing the initial modal
        confirmation dialog."""
        ok_to_proceed = self._confirm_uninstallation()
        if not ok_to_proceed:
            self._finalize()
            return

        self.dialog_timer.start()
        self._run_uninstaller()

    def _confirm_uninstallation(self) -> bool:
        """Present a modal dialog asking the user if they really want to uninstall. Returns True to
        continue with the uninstallation, or False to stop the process."""
        confirm = MessageDialog.show_modal(
            MessageDialog.DialogType.QUESTION,
            "AddonManager_ConfirmUninstallDialog",
            translate("AddonsInstaller", "Confirm remove"),
            translate("AddonsInstaller", "Are you sure you want to uninstall {}?").format(
                self.addon_to_remove.display_name
            ),
            QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.Cancel,
        )
        return confirm == QtWidgets.QMessageBox.Yes

    def _show_progress_dialog(self):
        self.progress_dialog = QtWidgets.QMessageBox(
            QtWidgets.QMessageBox.NoIcon,
            translate("AddonsInstaller", "Removing Addon"),
            translate("AddonsInstaller", "Removing {}").format(self.addon_to_remove.display_name)
            + "...",
            QtWidgets.QMessageBox.Cancel,
            parent=utils.get_main_am_window(),
        )
        self.progress_dialog.setObjectName("AddonManager_RemovingAddonDialog")
        self.progress_dialog.rejected.connect(self._cancel_removal)
        self.progress_dialog.show()

    def _run_uninstaller(self):
        self.worker_thread.start()

    def _cancel_removal(self):
        """Ask the QThread to interrupt. Probably has no effect, most of the work is in a single OS
        call."""
        self.worker_thread.requestInterruption()

    def _succeeded(self, addon):
        """Callback for successful removal"""
        self.dialog_timer.stop()
        if self.progress_dialog:
            self.progress_dialog.hide()
        MessageDialog.show_modal(
            MessageDialog.DialogType.INFO,
            "AddonManager_UninstallCompleteDialog",
            translate("AddonsInstaller", "Uninstall complete"),
            translate("AddonInstaller", "Finished removing {}").format(addon.display_name),
            QtWidgets.QMessageBox.Ok,
        )
        self._finalize()

    def _failed(self, addon, message):
        """Callback for failed or partially failed removal"""
        self.dialog_timer.stop()
        if self.progress_dialog:
            self.progress_dialog.hide()
        MessageDialog.show_modal(
            MessageDialog.DialogType.ERROR,
            "AddonManager_UninstallFailedDialog",
            translate("AddonsInstaller", "Uninstall failed"),
            translate("AddonInstaller", "Failed to remove some files") + ":\n" + message,
            QtWidgets.QMessageBox.Ok,
        )
        self._finalize()

    def _finalize(self):
        """Clean up and emit the finished signal"""
        if self.worker_thread.isRunning():
            self.worker_thread.requestInterruption()
            self.worker_thread.quit()
            self.worker_thread.wait(500)
        self.finished.emit()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import glob
import os
import pathlib
from collections import defaultdict
from typing import Optional


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate

PreferencesGroup = "User parameter:BaseApp/Preferences/Mod/CAM"

DefaultFilePath = "DefaultFilePath"
DefaultJobTemplate = "DefaultJobTemplate"
DefaultStockTemplate = "DefaultStockTemplate"
DefaultTaskPanelLayout = "DefaultTaskPanelLayout"

PostProcessorDefault = "PostProcessorDefault"
PostProcessorDefaultArgs = "PostProcessorDefaultArgs"
PostProcessorBlacklist = "PostProcessorBlacklist"
PostProcessorOutputFile = "PostProcessorOutputFile"
PostProcessorOutputPolicy = "PostProcessorOutputPolicy"

ToolGroup = PreferencesGroup + "/Tools"
ToolPath = "ToolPath"
LastToolLibrary = "LastToolLibrary"
LastToolLibrarySortKey = "LastToolLibrarySortKey"

# Linear tolerance to use when generating Paths, eg when tessellating geometry
GeometryTolerance = "GeometryTolerance"
LibAreaCurveAccuracy = "LibAreaCurveAccuracy"

WarningSuppressRapidSpeeds = "WarningSuppressRapidSpeeds"
WarningSuppressAllSpeeds = "WarningSuppressAllSpeeds"
WarningSuppressSelectionMode = "WarningSuppressSelectionMode"
WarningSuppressOpenCamLib = "WarningSuppressOpenCamLib"
WarningSuppressVelocity = "WarningSuppressVelocity"
EnableExperimentalFeatures = "EnableExperimentalFeatures"
EnableAdvancedOCLFeatures = "EnableAdvancedOCLFeatures"


_observers = defaultdict(list)  # maps group name to callback functions


def _add_group_observer(group, callback):
    """Add an observer for any changes on the given parameter group"""
    _observers[group].append(callback)


def _emit_change(group, *args):
    for cb in _observers[group]:
        cb(group, *args)


def preferences():
    return FreeCAD.ParamGet(PreferencesGroup)


def tool_preferences():
    return FreeCAD.ParamGet(ToolGroup)


def addToolPreferenceObserver(callback):
    _add_group_observer(ToolGroup, callback)


def pathPostSourcePath():
    return os.path.join(FreeCAD.getHomePath(), "Mod/CAM/Path/Post/")


def getBuiltinAssetPath() -> pathlib.Path:
    home = pathlib.Path(FreeCAD.getHomePath())
    return home / "Mod" / "CAM" / "Tools"


def getBuiltinLibraryPath() -> pathlib.Path:
    return getBuiltinAssetPath() / "Library"


def getBuiltinShapePath() -> pathlib.Path:
    return getBuiltinAssetPath() / "Shape"


def getBuiltinToolBitPath() -> pathlib.Path:
    return getBuiltinAssetPath() / "Bit"


def getDefaultAssetPath() -> Path:
    data_dir = pathlib.Path(FreeCAD.getUserAppDataDir())
    asset_path = data_dir / "CamAssets"
    asset_path.mkdir(parents=True, exist_ok=True)
    return asset_path


def getAssetPath() -> pathlib.Path:
    pref = tool_preferences()

    # Check if we have a CamAssets path already set
    cam_assets_path = pref.GetString(ToolPath, "")
    if cam_assets_path:
        # Use mostRecentConfigFromBase to get the most recent versioned path
        most_recent_path = FreeCAD.ApplicationDirectories.mostRecentConfigFromBase(cam_assets_path)
        return pathlib.Path(most_recent_path)

    # Migration: Check for legacy DefaultFilePath and use it for CamAssets
    legacy_path = defaultFilePath()
    if legacy_path:
        legacy_path_obj = pathlib.Path(legacy_path)
        if legacy_path_obj.exists() and legacy_path_obj.is_dir():
            # Migrate: Set the legacy path as the new CamAssets path
            setAssetPath(legacy_path_obj)
            # Return the most recent version of the legacy path
            most_recent_legacy = FreeCAD.ApplicationDirectories.mostRecentConfigFromBase(
                str(legacy_path_obj)
            )
            return pathlib.Path(most_recent_legacy)

    # Fallback to default if no legacy path found
    default = getDefaultAssetPath()
    # Return the most recent version of the default path
    most_recent_default = FreeCAD.ApplicationDirectories.mostRecentConfigFromBase(str(default))
    return pathlib.Path(most_recent_default)


def setAssetPath(path: pathlib.Path):
    Path.Log.debug(f"Setting asset path to {path}")
    assert path.is_dir(), f"Cannot put a non-initialized asset directory into preferences: {path}"
    pref = tool_preferences()
    current_path = pref.GetString(ToolPath, "")
    if str(path) == current_path:
        return
    pref.SetString(ToolPath, str(path))
    _emit_change(ToolGroup, ToolPath, path)


def getToolBitPath() -> pathlib.Path:
    return getAssetPath() / "Tools" / "Bit"


def getTemplateDirectory() -> pathlib.Path:
    """Returns the directory where job templates should be saved."""
    template_path = getAssetPath() / "Templates"
    template_path.mkdir(parents=True, exist_ok=True)
    return template_path


def getLastToolLibrary() -> Optional[str]:
    pref = tool_preferences()
    return pref.GetString(LastToolLibrary) or None


def setLastToolLibrary(name: str):
    assert isinstance(name, str), f"Library name '{name}' is not a string"
    pref = tool_preferences()
    pref.SetString(LastToolLibrary, name)


def getLastToolLibrarySortKey() -> Optional[str]:
    pref = tool_preferences()
    return pref.GetString(LastToolLibrarySortKey) or None


def setLastToolLibrarySortKey(name: str):
    pref = tool_preferences()
    pref.SetString(LastToolLibrarySortKey, name)


def allAvailablePostProcessors():
    allposts = []
    for path in searchPathsPost():
        posts = [
            str(os.path.split(os.path.splitext(p)[0])[1][:-5])
            for p in glob.glob(path + "/*_post.py")
        ]
        allposts.extend(posts)
    allposts.sort()
    return allposts


def allEnabledPostProcessors(include=None):
    blacklist = postProcessorBlacklist()
    enabled = [
        processor for processor in allAvailablePostProcessors() if processor not in blacklist
    ]
    if include:
        postlist = list(set(include + enabled))
        postlist.sort()
        return postlist
    return enabled


def defaultPostProcessor():
    pref = preferences()
    return pref.GetString(PostProcessorDefault, "")


def defaultPostProcessorArgs():
    pref = preferences()
    return pref.GetString(PostProcessorDefaultArgs, "")


def defaultGeometryTolerance():
    return preferences().GetFloat(GeometryTolerance, 0.01)


def defaultLibAreaCurveAccuracy():
    return preferences().GetFloat(LibAreaCurveAccuracy, 0.01)


def defaultFilePath():
    return preferences().GetString(DefaultFilePath)


def filePath():
    path = defaultFilePath()
    if not path:
        path = str(getAssetPath())
    return path


def macroFilePath():
    grp = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Macro")
    return grp.GetString("MacroPath", FreeCAD.getUserMacroDir())


def searchPaths():
    paths = []
    # Add new CamAssets/Templates directory first (highest priority)
    paths.append(str(getTemplateDirectory()))
    # Add legacy locations for backward compatibility
    p = defaultFilePath()
    if p:
        paths.append(p)
    paths.append(macroFilePath())
    return paths


def searchPathsPost():
    paths = []
    p = defaultFilePath()
    if p:
        paths.append(p)
    paths.append(macroFilePath())
    paths.append(os.path.join(pathPostSourcePath(), "scripts/"))
    paths.append(pathPostSourcePath())
    return paths


def defaultJobTemplate():
    template = preferences().GetString(DefaultJobTemplate)
    if "xml" not in template:
        return template
    return ""


def setJobDefaults(jobTemplate, geometryTolerance, curveAccuracy):
    Path.Log.track("(%s, %s, %s)" % (jobTemplate, geometryTolerance, curveAccuracy))
    pref = preferences()
    pref.SetString(DefaultJobTemplate, jobTemplate)
    pref.SetFloat(GeometryTolerance, geometryTolerance)
    pref.SetFloat(LibAreaCurveAccuracy, curveAccuracy)


def postProcessorBlacklist():
    pref = preferences()
    blacklist = pref.GetString(PostProcessorBlacklist, "")
    if not blacklist:
        return []
    return eval(blacklist)


def setPostProcessorDefaults(processor, args, blacklist):
    pref = preferences()
    pref.SetString(PostProcessorDefault, processor)
    pref.SetString(PostProcessorDefaultArgs, args)
    pref.SetString(PostProcessorBlacklist, "%s" % (blacklist))


def setOutputFileDefaults(fileName, policy):
    pref = preferences()
    pref.SetString(PostProcessorOutputFile, fileName)
    pref.SetString(PostProcessorOutputPolicy, policy)


def defaultOutputFile():
    pref = preferences()
    return pref.GetString(PostProcessorOutputFile, "")


def defaultOutputPolicy():
    pref = preferences()
    return pref.GetString(PostProcessorOutputPolicy, "")


def defaultStockTemplate():
    return preferences().GetString(DefaultStockTemplate, "")


def setDefaultStockTemplate(template):
    preferences().SetString(DefaultStockTemplate, template)


def defaultTaskPanelLayout():
    return preferences().GetInt(DefaultTaskPanelLayout, 0)


def setDefaultTaskPanelLayout(style):
    preferences().SetInt(DefaultTaskPanelLayout, style)


def advancedOCLFeaturesEnabled():
    return preferences().GetBool(EnableAdvancedOCLFeatures, False)


def experimentalFeaturesEnabled():
    return preferences().GetBool(EnableExperimentalFeatures, False)


def suppressAllSpeedsWarning():
    return preferences().GetBool(WarningSuppressAllSpeeds, True)


def suppressRapidSpeedsWarning(user=True):
    return (user and suppressAllSpeedsWarning()) or preferences().GetBool(
        WarningSuppressRapidSpeeds, True
    )


def suppressSelectionModeWarning():
    return preferences().GetBool(WarningSuppressSelectionMode, True)


def suppressOpenCamLibWarning():
    return preferences().GetBool(WarningSuppressOpenCamLib, True)


def suppressVelocity():
    return preferences().GetBool(WarningSuppressVelocity, False)


def setPreferencesAdvanced(ocl, warnSpeeds, warnRapids, warnModes, warnOCL, warnVelocity):
    preferences().SetBool(EnableAdvancedOCLFeatures, ocl)
    preferences().SetBool(WarningSuppressAllSpeeds, warnSpeeds)
    preferences().SetBool(WarningSuppressRapidSpeeds, warnRapids)
    preferences().SetBool(WarningSuppressSelectionMode, warnModes)
    preferences().SetBool(WarningSuppressOpenCamLib, warnOCL)
    preferences().SetBool(WarningSuppressVelocity, warnVelocity)



# SPDX-License-Identifier: LGPL-2.1-or-later

import unittest
from Path.Tool.assets.uri import AssetUri


class TestPathToolAssetUri(unittest.TestCase):
    """
    Test suite for the AssetUri utility class.
    """

    def test_uri_parsing_full(self):
        uri_string = "remote://asset_id/version?" "param1=value1&param2=value2"
        uri = AssetUri(uri_string)
        self.assertEqual(uri.asset_type, "remote")
        self.assertEqual(uri.asset_id, "asset_id")
        self.assertEqual(uri.version, "version")
        self.assertEqual(uri.params, {"param1": ["value1"], "param2": ["value2"]})
        self.assertEqual(str(uri), uri_string)
        self.assertEqual(repr(uri), f"AssetUri('{uri_string}')")

    def test_uri_parsing_local(self):
        uri_string = "local://id/2?param=value"
        uri = AssetUri(uri_string)
        self.assertEqual(uri.asset_type, "local")
        self.assertEqual(uri.asset_id, "id")
        self.assertEqual(uri.version, "2")
        self.assertEqual(uri.params, {"param": ["value"]})
        self.assertEqual(str(uri), uri_string)
        self.assertEqual(repr(uri), f"AssetUri('{uri_string}')")

    def test_uri_parsing_no_params(self):
        uri_string = "file://asset_id/1"
        uri = AssetUri(uri_string)
        self.assertEqual(uri.asset_type, "file")
        self.assertEqual(uri.asset_id, "asset_id")
        self.assertEqual(uri.version, "1")
        self.assertEqual(uri.params, {})
        self.assertEqual(str(uri), uri_string)
        self.assertEqual(repr(uri), f"AssetUri('{uri_string}')")

    def test_uri_version_missing(self):
        uri_string = "foo://asset"
        uri = AssetUri(uri_string)
        self.assertEqual(uri.asset_type, "foo")
        self.assertEqual(uri.asset_id, "asset")
        self.assertIsNone(uri.version)
        self.assertEqual(uri.params, {})
        self.assertEqual(str(uri), uri_string)

    def test_uri_parsing_with_version(self):
        """
        Test parsing a URI string with asset_type, asset_id, and version.
        """
        uri_string = "test_type://test_id/1"
        uri = AssetUri(uri_string)
        self.assertEqual(uri.asset_type, "test_type")
        self.assertEqual(uri.asset_id, "test_id")
        self.assertEqual(uri.version, "1")
        self.assertEqual(uri.params, {})
        self.assertEqual(str(uri), uri_string)
        self.assertEqual(repr(uri), f"AssetUri('{uri_string}')")

    def test_uri_build_full(self):
        expected_uri_string = "local://asset_id/version?param1=value1"
        uri = AssetUri.build(
            asset_type="local", asset_id="asset_id", version="version", params={"param1": "value1"}
        )
        self.assertEqual(str(uri), expected_uri_string)
        self.assertEqual(uri.asset_type, "local")
        self.assertEqual(uri.asset_id, "asset_id")
        self.assertEqual(uri.version, "version")
        self.assertEqual(uri.params, {"param1": ["value1"]})  # parse_qs always returns list

    def test_uri_build_latest_version_no_params(self):
        expected_uri_string = "remote://id/latest"
        uri = AssetUri.build(asset_type="remote", asset_id="id", version="latest")
        self.assertEqual(str(uri), expected_uri_string)
        self.assertEqual(uri.asset_type, "remote")
        self.assertEqual(uri.asset_id, "id")
        self.assertEqual(uri.version, "latest")
        self.assertEqual(uri.params, {})

    def test_uri_equality(self):
        uri1 = AssetUri("local://asset/version")
        uri2 = AssetUri("local://asset/version")
        uri3 = AssetUri("local://asset/another_version")
        self.assertEqual(uri1, uri2)
        self.assertNotEqual(uri1, uri3)
        self.assertNotEqual(uri1, "not a uri")

    def test_uri_parsing_invalid_path_structure(self):
        """
        Test that parsing a URI string with an invalid path structure
        (more than one component) raises a ValueError.
        """
        uri_string = "local://foo/bar/1"
        with self.assertRaisesRegex(ValueError, "Invalid URI path structure:"):
            AssetUri(uri_string)


if __name__ == "__main__":
    unittest.main()



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 FreeCAD Developers                                 *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the viewprovider code for the WorkingPlaneProxy object."""
## @package view_wpproxy
# \ingroup draftviewproviders
# \brief Provides the viewprovider code for the WorkingPlaneProxy object.

## \addtogroup draftviewproviders
# @{
import pivy.coin as coin
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
from draftutils import gui_utils
from draftutils import params


class ViewProviderWorkingPlaneProxy:
    """A View Provider for working plane proxies"""

    def __init__(self, vobj):
        # ViewData: 0,1,2: position; 3,4,5,6: rotation; 7: near dist; 8: far dist, 9:aspect ratio;
        # 10: focal dist; 11: height (ortho) or height angle (persp); 12: ortho (0) or persp (1)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The display length of this section plane")
        vobj.addProperty("App::PropertyLength", "DisplaySize", "Draft", _tip, locked=True)

        _tip = QT_TRANSLATE_NOOP("App::Property", "The size of the arrows of this section plane")
        vobj.addProperty("App::PropertyLength", "ArrowSize", "Draft", _tip, locked=True)

        vobj.addProperty("App::PropertyPercent", "Transparency", "Base", "", locked=True)

        vobj.addProperty("App::PropertyFloat", "LineWidth", "Base", "", locked=True)

        vobj.addProperty("App::PropertyColor", "LineColor", "Base", "", locked=True)

        vobj.addProperty("App::PropertyFloatList", "ViewData", "Base", "", locked=True)

        vobj.addProperty("App::PropertyBool", "RestoreView", "Base", "", locked=True)

        vobj.addProperty("App::PropertyMap", "VisibilityMap", "Base", "", locked=True)

        vobj.addProperty("App::PropertyBool", "RestoreState", "Base", "", locked=True)

        vobj.DisplaySize = 100
        vobj.ArrowSize = 5
        vobj.Transparency = 70
        vobj.LineWidth = 1
        vobj.LineColor = params.get_param("gridColor") | 0x000000FF

        vobj.Proxy = self
        vobj.RestoreView = True
        vobj.RestoreState = True

    def getIcon(self):
        return ":/icons/Draft_SelectPlane.svg"

    def claimChildren(self):
        return []

    def doubleClicked(self, vobj):
        Gui.runCommand("Draft_SelectPlane")
        return True

    def setupContextMenu(self, vobj, menu):
        action1 = QtGui.QAction(
            QtGui.QIcon(":/icons/Draft_SelectPlane.svg"), "Save Camera Position", menu
        )
        QtCore.QObject.connect(action1, QtCore.SIGNAL("triggered()"), self.writeCamera)
        menu.addAction(action1)
        action2 = QtGui.QAction(
            QtGui.QIcon(":/icons/Draft_SelectPlane.svg"), "Save Visibility of Objects", menu
        )
        QtCore.QObject.connect(action2, QtCore.SIGNAL("triggered()"), self.writeState)
        menu.addAction(action2)

    def writeCamera(self):
        if hasattr(self, "Object"):
            n = Gui.ActiveDocument.ActiveView.getCameraNode()
            App.Console.PrintMessage(QT_TRANSLATE_NOOP("Draft", "Writing camera position") + "\n")
            cdata = list(n.position.getValue().getValue())
            cdata.extend(list(n.orientation.getValue().getValue()))
            cdata.append(n.nearDistance.getValue())
            cdata.append(n.farDistance.getValue())
            cdata.append(n.aspectRatio.getValue())
            cdata.append(n.focalDistance.getValue())
            if isinstance(n, coin.SoOrthographicCamera):
                cdata.append(n.height.getValue())
                cdata.append(0.0)  # orthograhic camera
            elif isinstance(n, coin.SoPerspectiveCamera):
                cdata.append(n.heightAngle.getValue())
                cdata.append(1.0)  # perspective camera
            self.Object.ViewObject.ViewData = cdata

    def writeState(self):
        if hasattr(self, "Object"):
            App.Console.PrintMessage(
                QT_TRANSLATE_NOOP("Draft", "Writing objects shown/hidden state") + "\n"
            )
            vis = {}
            for o in App.ActiveDocument.Objects:
                if o.ViewObject:
                    vis[o.Name] = str(o.ViewObject.Visibility)
            self.Object.ViewObject.VisibilityMap = vis

    def attach(self, vobj):
        self.clip = None
        self.main_transform = gui_utils.find_coin_node(vobj.RootNode, coin.SoTransform)
        self.mat1 = coin.SoMaterial()
        self.mat2 = coin.SoMaterial()
        self.fcoords = coin.SoCoordinate3()
        fs = coin.SoIndexedFaceSet()
        fs.coordIndex.setValues(0, 7, [0, 1, 2, -1, 0, 2, 3])
        self.drawstyle = coin.SoDrawStyle()
        self.drawstyle.style = coin.SoDrawStyle.LINES
        self.lcoords = coin.SoCoordinate3()
        import PartGui  # Required for "SoBrepEdgeSet" (because a WorkingPlaneProxy is not a Part::FeaturePython object).

        ls = coin.SoType.fromName("SoBrepEdgeSet").createInstance()
        # fmt: off
        ls.coordIndex.setValues(
            0,
            28,
            [
                0, 1, -1,
                2, 3, 4, 5, -1,
                6, 7, -1,
                8, 9, 10, 11, -1,
                12, 13, -1,
                14, 15, 16, 17, -1,
                18, 19, 20, 21
            ],
        )
        # fmt: on
        sep = coin.SoSeparator()
        psep = coin.SoSeparator()
        fsep = coin.SoSeparator()
        fsep.addChild(self.mat2)
        fsep.addChild(self.fcoords)
        fsep.addChild(fs)
        psep.addChild(self.mat1)
        psep.addChild(self.drawstyle)
        psep.addChild(self.lcoords)
        psep.addChild(ls)
        sep.addChild(fsep)
        sep.addChild(psep)
        vobj.addDisplayMode(sep, "Default")
        self.onChanged(vobj, "DisplaySize")
        self.onChanged(vobj, "LineColor")
        self.onChanged(vobj, "Transparency")
        self.Object = vobj.Object

    def getDisplayModes(self, vobj):
        return ["Default"]

    def getDefaultDisplayMode(self):
        return "Default"

    def setDisplayMode(self, mode):
        return mode

    def updateData(self, obj, prop):
        if prop in ["Placement"]:
            self.onChanged(obj.ViewObject, "DisplaySize")
        return

    def onChanged(self, vobj, prop):
        if prop == "LineColor":
            l = vobj.LineColor
            self.mat1.diffuseColor.setValue([l[0], l[1], l[2]])
            self.mat2.diffuseColor.setValue([l[0], l[1], l[2]])

        elif prop == "Transparency":
            if hasattr(vobj, "Transparency"):
                self.mat2.transparency.setValue(vobj.Transparency / 100.0)

        elif prop in ["DisplaySize", "ArrowSize"]:
            if hasattr(vobj, "DisplaySize"):
                l = vobj.DisplaySize.Value / 2
            else:
                l = 1
            if hasattr(vobj, "ArrowSize"):
                l1 = vobj.ArrowSize.Value if vobj.ArrowSize.Value > 0 else 0.1
            else:
                l1 = 0.1
            l2 = l1 / 3

            pl = vobj.Object.Placement
            self.main_transform.translation.setValue(pl.Base)
            self.main_transform.rotation = coin.SbRotation(pl.Rotation.Q)
            verts = []
            fverts = []

            fverts.append(App.Vector(-l, -l, 0))
            fverts.append(App.Vector(l, -l, 0))
            fverts.append(App.Vector(l, l, 0))
            fverts.append(App.Vector(-l, l, 0))

            verts.append(App.Vector(0, 0, 0))
            verts.append(App.Vector(l - l1, 0, 0))
            verts.append(App.Vector(l - l1, l2, 0))
            verts.append(App.Vector(l, 0, 0))
            verts.append(App.Vector(l - l1, -l2, 0))
            verts.append(App.Vector(l - l1, l2, 0))

            verts.append(App.Vector(0, 0, 0))
            verts.append(App.Vector(0, l - l1, 0))
            verts.append(App.Vector(-l2, l - l1, 0))
            verts.append(App.Vector(0, l, 0))
            verts.append(App.Vector(l2, l - l1, 0))
            verts.append(App.Vector(-l2, l - l1, 0))

            verts.append(App.Vector(0, 0, 0))
            verts.append(App.Vector(0, 0, l - l1))
            verts.append(App.Vector(-l2, 0, l - l1))
            verts.append(App.Vector(0, 0, l))
            verts.append(App.Vector(l2, 0, l - l1))
            verts.append(App.Vector(-l2, 0, l - l1))
            verts.append(App.Vector(0, -l2, l - l1))
            verts.append(App.Vector(0, 0, l))
            verts.append(App.Vector(0, l2, l - l1))
            verts.append(App.Vector(0, -l2, l - l1))

            self.lcoords.point.setValues(verts)
            self.fcoords.point.setValues(fverts)

        elif prop == "LineWidth":
            self.drawstyle.lineWidth = vobj.LineWidth
        return

    def dumps(self):
        return None

    def loads(self, state):
        return None


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2011 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides the Snapper class to define the snapping tools and modes.

This module provides tools to handle point snapping and
everything that goes with it (toolbar buttons, cursor icons, etc.).
It also creates the Draft grid, which is actually a tracker
defined by `gui_trackers.gridTracker`.
"""
## @package gui_snapper
#  \ingroup draftguitools
#  \brief Provides the Snapper class to define the snapping tools and modes.
#
#  This module provides tools to handle point snapping and
#  everything that goes with it (toolbar buttons, cursor icons, etc.).

## \addtogroup draftguitools
# @{
import collections as coll
import inspect
import itertools
import math
import pivy.coin as coin
import PySide.QtCore as QtCore
import PySide.QtGui as QtGui
import PySide.QtWidgets as QtWidgets

import FreeCAD as App
import FreeCADGui as Gui
import Part
import Draft
import DraftVecUtils
import DraftGeomUtils
import WorkingPlane
from draftguitools import gui_trackers as trackers
from draftutils import gui_utils
from draftutils import params
from draftutils.init_tools import get_draft_snap_commands
from draftutils.messages import _wrn
from draftutils.translate import translate

__title__ = "FreeCAD Draft Snap tools"
__author__ = "Yorik van Havre"
__url__ = "https://www.freecad.org"

UNSNAPPABLES = ("Image::ImagePlane",)


class Snapper:
    """Classes to manage snapping in Draft and Arch.

    The Snapper objects contains all the functionality used by draft
    and arch module to manage object snapping. It is responsible for
    finding snap points and displaying snap markers. Usually You
    only need to invoke it's snap() function, all the rest is taken
    care of.

    3 functions are useful for the scriptwriter: snap(), constrain()
    or getPoint() which is an all-in-one combo.

    The individual snapToXXX() functions return a snap definition in
    the form [real_point,marker_type,visual_point], and are not
    meant to be used directly, they are all called when necessary by
    the general snap() function.

    The Snapper lives inside Gui once the Draft module has been
    loaded.

    """

    def __init__(self):
        self.activeview = None
        self.lastObj = []
        self.lastObjSubelements = []
        self.radius = 0
        self.constraintAxis = None
        self.basepoint = None
        self.affinity = None
        self.mask = None
        self.cursorMode = None
        self.cursorQt = None
        self.maxEdges = params.get_param("maxSnapEdges")

        # we still have no 3D view when the draft module initializes
        self.tracker = None
        self.extLine = None
        self.grid = None
        self.constrainLine = None
        self.trackLine = None
        self.extLine2 = None
        self.radiusTracker = None
        self.dim1 = None
        self.dim2 = None
        self.snapInfo = None
        self.lastExtensions = []
        # the trackers are stored in lists because there can be several views,
        # each with its own set
        # view, grid, snap, extline, radius, dim1, dim2, trackLine,
        # extline2, crosstrackers
        self.trackers = [[], [], [], [], [], [], [], [], [], []]
        self.polarAngles = [90, 45]
        self.selectMode = False
        self.holdTracker = None
        self.holdPoints = []
        self.running = False
        self.callbackClick = None
        self.callbackMove = None
        self.snapObjectIndex = 0

        # snap keys, it's important that they are in this order for
        # saving in preferences and for properly restoring the toolbar
        # fmt: off
        self.snaps = ['Lock',           # 0
                      'Near',           # 1 former "passive" snap
                      'Extension',      # 2
                      'Parallel',       # 3
                      'Grid',           # 4
                      "Endpoint",       # 5
                      'Midpoint',       # 6
                      'Perpendicular',  # 7
                      'Angle',          # 8
                      'Center',         # 9
                      'Ortho',          # 10
                      'Intersection',   # 11
                      'Special',        # 12
                      'Dimensions',     # 13
                      'WorkingPlane'    # 14
                     ]
        # fmt: on

        self.init_active_snaps()
        self.set_snap_style()

        self.cursors = coll.OrderedDict(
            [
                ("passive", ":/icons/Draft_Snap_Near.svg"),
                ("extension", ":/icons/Draft_Snap_Extension.svg"),
                ("parallel", ":/icons/Draft_Snap_Parallel.svg"),
                ("grid", ":/icons/Draft_Snap_Grid.svg"),
                ("endpoint", ":/icons/Draft_Snap_Endpoint.svg"),
                ("midpoint", ":/icons/Draft_Snap_Midpoint.svg"),
                ("perpendicular", ":/icons/Draft_Snap_Perpendicular.svg"),
                ("angle", ":/icons/Draft_Snap_Angle.svg"),
                ("center", ":/icons/Draft_Snap_Center.svg"),
                ("ortho", ":/icons/Draft_Snap_Ortho.svg"),
                ("intersection", ":/icons/Draft_Snap_Intersection.svg"),
                ("special", ":/icons/Draft_Snap_Special.svg"),
            ]
        )

    def _get_wp(self):
        # update=False is required, without it WorkingPlane.get_working_plane()
        # is too slow for this function which gets called repeatedly when moving
        # the mouse
        # See: https://github.com/FreeCAD/FreeCAD/issues/24013
        return WorkingPlane.get_working_plane(update=False)

    def init_active_snaps(self):
        """
        set self.active_snaps according to user prefs
        """
        self.active_snaps = []
        snap_modes = params.get_param("snapModes")
        i = 0
        for snap in snap_modes:
            if bool(int(snap)):
                self.active_snaps.append(self.snaps[i])
            i += 1

    def set_snap_style(self):
        self.snapStyle = params.get_param("snapStyle")
        if self.snapStyle:
            self.mk = coll.OrderedDict(
                [
                    ("passive", "SQUARE_LINE"),
                    ("extension", "SQUARE_LINE"),
                    ("parallel", "SQUARE_LINE"),
                    ("grid", "SQUARE_FILLED"),
                    ("endpoint", "SQUARE_FILLED"),
                    ("midpoint", "SQUARE_FILLED"),
                    ("perpendicular", "SQUARE_FILLED"),
                    ("angle", "SQUARE_FILLED"),
                    ("center", "SQUARE_FILLED"),
                    ("ortho", "SQUARE_FILLED"),
                    ("intersection", "SQUARE_FILLED"),
                    ("special", "SQUARE_FILLED"),
                ]
            )
        else:
            self.mk = coll.OrderedDict(
                [
                    ("passive", "CIRCLE_LINE"),
                    ("extension", "CIRCLE_LINE"),
                    ("parallel", "CIRCLE_LINE"),
                    ("grid", "CIRCLE_LINE"),
                    ("endpoint", "CIRCLE_FILLED"),
                    ("midpoint", "DIAMOND_FILLED"),
                    ("perpendicular", "CIRCLE_FILLED"),
                    ("angle", "DIAMOND_FILLED"),
                    ("center", "CIRCLE_FILLED"),
                    ("ortho", "CIRCLE_FILLED"),
                    ("intersection", "CIRCLE_FILLED"),
                    ("special", "CIRCLE_FILLED"),
                ]
            )

    def cstr(self, lastpoint, constrain, point):
        """Return constraints if needed."""
        if constrain or self.mask:
            fpt = self.constrain(point, lastpoint)
        else:
            self.unconstrain()
            fpt = point
        if self.radiusTracker:
            self.radiusTracker.update(fpt)
        return fpt

    def snap(self, screenpos, lastpoint=None, active=True, constrain=False, noTracker=False):
        """Return a snapped point from the given (x, y) screen position.

        snap(screenpos,lastpoint=None,active=True,constrain=False,
        noTracker=False): returns a snapped point from the given
        (x,y) screenpos (the position of the mouse cursor), active is to
        activate active point snapping or not (passive),
        lastpoint is an optional other point used to draw an
        imaginary segment and get additional snap locations. Constrain can
        be True to constrain the point against the closest working plane axis.
        Screenpos can be a list, a tuple or a coin.SbVec2s object.
        If noTracker is True, the tracking line is not displayed.
        """
        if self.running:
            # do not allow concurrent runs
            return None

        self.running = True

        self.spoint = None

        if params.get_param("SnapBarShowOnlyDuringCommands"):
            toolbar = self.get_snap_toolbar()
            if toolbar:
                toolbar.show()

        self.snapInfo = None

        # Type conversion if needed
        if isinstance(screenpos, list):
            screenpos = tuple(screenpos)
        elif isinstance(screenpos, coin.SbVec2s):
            screenpos = tuple(screenpos.getValue())
        elif not isinstance(screenpos, tuple):
            _wrn("Snap needs valid screen position (list, tuple or sbvec2s)")
            self.running = False
            return None

        # Setup trackers if needed
        self.setTrackers()

        # Get current snap radius
        self.radius = self.getScreenDist(params.get_param("snapRange"), screenpos)
        if self.radiusTracker:
            self.radiusTracker.update(self.radius)
            self.radiusTracker.off()

        # Activate snap
        if params.get_param("alwaysSnap"):
            active = True

        self.setCursor("passive")
        if self.tracker:
            self.tracker.off()
        if self.extLine2:
            self.extLine2.off()
        if self.extLine:
            self.extLine.off()
        if self.trackLine:
            self.trackLine.off()
        if self.dim1:
            self.dim1.off()
        if self.dim2:
            self.dim2.off()

        point = self.getApparentPoint(screenpos[0], screenpos[1])

        # Set up a track line if we got a last point
        if lastpoint and self.trackLine:
            self.trackLine.p1(lastpoint)

        # Check if parallel to one of the edges of the last objects
        # or to a polar direction
        eline = None
        if active:
            point, eline = self.snapToPolar(point, lastpoint)
            point, eline = self.snapToExtensions(point, lastpoint, constrain, eline)

        # Check if we have an object under the cursor and try to
        # snap to it
        _view = Draft.get3DView()
        objectsUnderCursor = _view.getObjectsInfo((screenpos[0], screenpos[1]))
        if objectsUnderCursor:
            if self.snapObjectIndex >= len(objectsUnderCursor):
                self.snapObjectIndex = 0
            self.snapInfo = objectsUnderCursor[self.snapObjectIndex]

        if self.snapInfo and "Component" in self.snapInfo:
            osnap = self.snapToObject(lastpoint, active, constrain, eline, point)
            if osnap:
                self.running = False
                return osnap

        # Nothing has been snapped.
        # Check for grid snap and ext crossings
        if active:
            epoint = self.snapToCrossExtensions(point)
            if epoint:
                point = epoint
            else:
                point = self.snapToGrid(point)
        fp = self.cstr(lastpoint, constrain, point)
        if self.trackLine and lastpoint and (not noTracker):
            self.trackLine.p2(fp)
            self.trackLine.setColor()
            self.trackLine.on()
        # Set the arch point tracking
        if lastpoint:
            self.setArchDims(lastpoint, fp)

        self.spoint = fp
        self.running = False
        return fp

    def cycleSnapObject(self):
        """Increase the index of the snap object by one."""
        self.snapObjectIndex += 1

    def snapToObject(self, lastpoint, active, constrain, eline, point):
        """Snap to an object."""

        if not active:
            return None

        parent = self.snapInfo.get("ParentObject", None)
        if parent:
            subname = self.snapInfo["SubName"]
            obj = parent.getSubObject(subname, retType=1)
        else:
            obj = App.ActiveDocument.getObject(self.snapInfo["Object"])
            parent = obj
            subname = self.snapInfo["Component"]

        if (
            not obj
            or Draft.getType(obj) in UNSNAPPABLES
            or not getattr(obj.ViewObject, "Selectable", True)
        ):
            # increase snapObjectIndex to find other objects under the cursor:
            self.snapObjectIndex += 1
            return None

        snaps = []
        point = App.Vector(self.snapInfo["x"], self.snapInfo["y"], self.snapInfo["z"])
        comp = self.snapInfo["Component"]
        shape = Part.getShape(parent, subname, needSubElement=True, noElementMap=True)

        if not shape.isNull():
            snaps.extend(self.snapToSpecials(obj, lastpoint, eline))

            if Draft.getType(obj) == "Polygon":
                # Special snapping for polygons: add the center
                snaps.extend(self.snapToPolygon(obj))

            elif Draft.getType(obj) == "BuildingPart" and self.isEnabled("Center"):
                # snap to the base placement of empty BuildingParts
                snaps.append([obj.Placement.Base, "center", self.toWP(obj.Placement.Base)])

            if (not self.maxEdges) or (len(shape.Edges) <= self.maxEdges):
                if "Edge" in comp:
                    # we are snapping to an edge
                    if shape.ShapeType == "Edge":
                        edge = shape
                        snaps.extend(self.snapToNear(edge, point))
                        snaps.extend(self.snapToEndpoints(edge))
                        snaps.extend(self.snapToMidpoint(edge))
                        snaps.extend(self.snapToPerpendicular(edge, lastpoint))
                        snaps.extend(self.snapToIntersection(edge))
                        snaps.extend(self.snapToElines(edge, eline))

                        et = DraftGeomUtils.geomType(edge)
                        if et == "Circle":
                            # the edge is an arc, we have extra options
                            snaps.extend(self.snapToAngles(edge))
                            snaps.extend(self.snapToCenter(edge))
                        elif et == "Ellipse":
                            # extra ellipse options
                            snaps.extend(self.snapToCenter(edge))
                elif "Face" in comp:
                    # we are snapping to a face
                    if shape.ShapeType == "Face":
                        face = shape
                        snaps.extend(self.snapToNearFace(face, point))
                        snaps.extend(self.snapToPerpendicularFace(face, lastpoint))
                        snaps.extend(self.snapToIntersection(face))
                        snaps.extend(self.snapToCenterFace(face))
                elif "Vertex" in comp:
                    # we are snapping to a vertex
                    if shape.ShapeType == "Vertex":
                        snaps.extend(self.snapToEndpoints(shape))
                else:
                    # `Catch-all` for other cases. Probably never executes
                    # as objects with a Shape typically have edges, faces
                    # or vertices.
                    snaps.extend(self.snapToNearUnprojected(point))

        elif Draft.getType(obj) in ("LinearDimension", "AngularDimension"):
            # for dimensions we snap to their 2 points:
            snaps.extend(self.snapToDim(obj))

        elif Draft.getType(obj) == "Axis":
            for edge in obj.Shape.Edges:
                snaps.extend(self.snapToEndpoints(edge))
                snaps.extend(self.snapToIntersection(edge))

        elif Draft.getType(obj).startswith("Mesh::"):
            snaps.extend(self.snapToNearUnprojected(point))
            snaps.extend(self.snapToEndpoints(obj.Mesh))

        elif Draft.getType(obj).startswith("Points::"):
            snaps.extend(self.snapToEndpoints(obj.Points, point))

        elif Draft.getType(obj) in ("WorkingPlaneProxy", "BuildingPart") and self.isEnabled(
            "Center"
        ):
            # snap to the center of WPProxies or to the base
            # placement of no empty BuildingParts
            snaps.append([obj.Placement.Base, "center", self.toWP(obj.Placement.Base)])

        elif Draft.getType(obj) == "SectionPlane":
            # snap to corners of section planes
            snaps.extend(self.snapToEndpoints(obj.Shape))

        # updating last objects list
        # objects must be added even if no snap has been found for the object
        # otherwise Intersection snap (for example) will not work
        if obj.Name in self.lastObj:
            self.lastObjSubelements.pop(self.lastObj.index(obj.Name))
            self.lastObj.remove(obj.Name)
        self.lastObj.append(obj.Name)
        self.lastObjSubelements.append(subname.split(".")[-1])
        if len(self.lastObj) > 8:
            self.lastObj = self.lastObj[-8:]
            self.lastObjSubelements = self.lastObjSubelements[-8:]

        if not snaps:
            return None

        # calculating the nearest snap point
        # a Near ("passive") snap point does not 'win' if a different snap point
        # is within snapRange of the cursor point (in screen coordinates)
        cursor_pt = App.Vector(self.snapInfo["x"], self.snapInfo["y"], self.snapInfo["z"])
        shortest_all = shortest_not_near = 1000000000000000000
        winner_all = winner_not_near = None
        for snap in snaps:
            if (not snap) or (snap[0] is None):
                pass
                # print("debug: Snapper: invalid snap point: ", snaps)
            else:
                dist = snap[0].sub(cursor_pt).Length
                if snap[1] != "passive":
                    if dist < shortest_not_near:
                        shortest_not_near = dist
                        winner_not_near = snap
                if dist < shortest_all:
                    shortest_all = dist
                    winner_all = snap

        if winner_not_near is None or shortest_not_near == shortest_all:
            winner = winner_all
        else:
            view = Draft.get3DView()
            # get screen points with pixel coordinates
            scr_win_not_near_pt = App.Vector(*view.getPointOnScreen(winner_not_near[0]), 0)
            scr_cursor_pt = App.Vector(*view.getPointOnScreen(cursor_pt), 0)
            if scr_win_not_near_pt.sub(scr_cursor_pt).Length <= params.get_param("snapRange"):
                winner = winner_not_near
            else:
                winner = winner_all

        fp = point
        if winner:
            # setting the cursors
            if self.tracker and not self.selectMode:
                self.tracker.setCoords(winner[2])
                self.tracker.setMarker(self.mk[winner[1]])
                self.tracker.on()
            # setting the trackline
            fp = self.cstr(lastpoint, constrain, winner[2])
            if self.trackLine and lastpoint:
                self.trackLine.p2(fp)
                self.trackLine.setColor()
                self.trackLine.on()
            # set the cursor
            self.setCursor(winner[1])

            # set the arch point tracking
            if lastpoint:
                self.setArchDims(lastpoint, fp)

        # return the final point
        self.spoint = fp
        return self.spoint

    def toWP(self, point):
        """Project the given point on the working plane, if needed."""
        if self.isEnabled("WorkingPlane"):
            return self._get_wp().project_point(point)
        return point

    def getApparentPoint(self, x, y):
        """Return a 3D point, projected on the current working plane."""
        view = Draft.get3DView()
        pt = view.getPoint(x, y)
        if self.mask != "z":
            if view.getCameraType() == "Perspective":
                camera = view.getCameraNode()
                p = camera.getField("position").getValue()
                dv = pt.sub(App.Vector(p[0], p[1], p[2]))
            else:
                dv = view.getViewDirection()
            return self._get_wp().project_point(pt, dv)
        return pt

    def snapToDim(self, obj):
        snaps = []
        if (
            self.isEnabled("Endpoint")
            and obj.ViewObject
            and hasattr(obj.ViewObject.Proxy, "p2")
            and hasattr(obj.ViewObject.Proxy, "p3")
        ):
            snaps.append([obj.ViewObject.Proxy.p2, "endpoint", self.toWP(obj.ViewObject.Proxy.p2)])
            snaps.append([obj.ViewObject.Proxy.p3, "endpoint", self.toWP(obj.ViewObject.Proxy.p3)])
        return snaps

    def snapToExtensions(self, point, last, constrain, eline):
        """Return a point snapped to extension or parallel line.

        The parallel line of the last object, if any.
        """
        tsnap = self.snapToHold(point)
        if tsnap:
            if self.tracker and not self.selectMode:
                self.tracker.setCoords(tsnap[2])
                self.tracker.setMarker(self.mk[tsnap[1]])
                self.tracker.on()
            if self.extLine:
                self.extLine.p1(tsnap[0])
                self.extLine.p2(tsnap[2])
                self.extLine.setColor()
                self.extLine.on()
            self.setCursor(tsnap[1])
            return tsnap[2], eline
        if self.isEnabled("Extension"):
            tsnap = self.snapToExtOrtho(last, constrain, eline)
            if tsnap:
                if (tsnap[0].sub(point)).Length < self.radius:
                    if self.tracker and not self.selectMode:
                        self.tracker.setCoords(tsnap[2])
                        self.tracker.setMarker(self.mk[tsnap[1]])
                        self.tracker.on()
                    if self.extLine:
                        self.extLine.p2(tsnap[2])
                        self.extLine.setColor()
                        self.extLine.on()
                    self.setCursor(tsnap[1])
                    return tsnap[2], eline
            else:
                tsnap = self.snapToExtPerpendicular(last)
                if tsnap:
                    if (tsnap[0].sub(point)).Length < self.radius:
                        if self.tracker and not self.selectMode:
                            self.tracker.setCoords(tsnap[2])
                            self.tracker.setMarker(self.mk[tsnap[1]])
                            self.tracker.on()
                        if self.extLine:
                            self.extLine.p2(tsnap[2])
                            self.extLine.setColor()
                            self.extLine.on()
                        self.setCursor(tsnap[1])
                        return tsnap[2], eline

        for o in self.lastObj:
            if self.isEnabled("Extension") or self.isEnabled("Parallel"):
                ob = App.ActiveDocument.getObject(o)
                if not ob:
                    continue
                if not ob.isDerivedFrom("Part::Feature"):
                    continue
                edges = ob.Shape.Edges
                if Draft.getType(ob) == "Wall":
                    for so in [ob] + ob.Additions:
                        if Draft.getType(so) == "Wall":
                            if so.Base:
                                edges.extend(so.Base.Shape.Edges)
                                edges.reverse()
                if (not self.maxEdges) or (len(edges) <= self.maxEdges):
                    for e in edges:
                        if DraftGeomUtils.geomType(e) != "Line":
                            continue
                        np = self.getPerpendicular(e, point)
                        if (np.sub(point)).Length < self.radius:
                            if self.isEnabled("Extension"):
                                if DraftGeomUtils.isPtOnEdge(np, e):
                                    continue
                                if np != e.Vertexes[0].Point:
                                    p0 = e.Vertexes[0].Point
                                    if self.tracker and not self.selectMode:
                                        self.tracker.setCoords(np)
                                        self.tracker.setMarker(self.mk["extension"])
                                        self.tracker.on()
                                    if self.extLine:
                                        self.extLine.p1(p0)
                                        self.extLine.p2(np)
                                        self.extLine.setColor()
                                        self.extLine.on()
                                    self.setCursor("extension")
                                    ne = Part.LineSegment(p0, np).toShape()
                                    # storing extension line for intersection calculations later
                                    if len(self.lastExtensions) == 0:
                                        self.lastExtensions.append(ne)
                                    elif len(self.lastExtensions) == 1:
                                        if not DraftGeomUtils.areColinear(
                                            ne, self.lastExtensions[0]
                                        ):
                                            self.lastExtensions.append(self.lastExtensions[0])
                                            self.lastExtensions[0] = ne
                                    else:
                                        if (
                                            not DraftGeomUtils.areColinear(
                                                ne, self.lastExtensions[0]
                                            )
                                        ) and (
                                            not DraftGeomUtils.areColinear(
                                                ne, self.lastExtensions[1]
                                            )
                                        ):
                                            self.lastExtensions[1] = self.lastExtensions[0]
                                            self.lastExtensions[0] = ne
                                    return np, ne
                        elif self.isEnabled("Parallel"):
                            if last:
                                ve = DraftGeomUtils.vec(e)
                                if not DraftVecUtils.isNull(ve):
                                    de = Part.LineSegment(last, last.add(ve)).toShape()
                                    np = self.getPerpendicular(de, point)
                                    if (np.sub(point)).Length < self.radius:
                                        if self.tracker and not self.selectMode:
                                            self.tracker.setCoords(np)
                                            self.tracker.setMarker(self.mk["parallel"])
                                            self.tracker.on()
                                        self.setCursor("parallel")
                                        return np, de
        return point, eline

    def snapToCrossExtensions(self, point):
        """Snap to the intersection of the last 2 extension lines."""
        if self.isEnabled("Extension"):
            if len(self.lastExtensions) == 2:
                np = DraftGeomUtils.findIntersection(
                    self.lastExtensions[0], self.lastExtensions[1], True, True
                )
                if np:
                    for p in np:
                        dv = point.sub(p)
                        if (self.radius == 0) or (dv.Length <= self.radius):
                            if self.tracker and not self.selectMode:
                                self.tracker.setCoords(p)
                                self.tracker.setMarker(self.mk["intersection"])
                                self.tracker.on()
                            self.setCursor("intersection")
                            if self.extLine and self.extLine2:
                                if DraftVecUtils.equals(
                                    self.extLine.p1(), self.lastExtensions[0].Vertexes[0].Point
                                ):
                                    p0 = self.lastExtensions[1].Vertexes[0].Point
                                else:
                                    p0 = self.lastExtensions[0].Vertexes[0].Point
                                self.extLine2.p1(p0)
                                self.extLine2.p2(p)
                                self.extLine.p2(p)
                                self.extLine.setColor()
                                self.extLine2.on()
                            return p
        return None

    def snapToPolar(self, point, last):
        """Snap to polar lines from the given point."""
        if self.isEnabled("Ortho") and (not self.mask):
            if last:
                vecs = []
                wp = self._get_wp()
                ax = [wp.u, wp.v, wp.axis]
                for a in self.polarAngles:
                    if a == 90:
                        vecs.extend([ax[0], ax[0].negative()])
                        vecs.extend([ax[1], ax[1].negative()])
                    else:
                        v = DraftVecUtils.rotate(ax[0], math.radians(a), ax[2])
                        vecs.extend([v, v.negative()])
                        v = DraftVecUtils.rotate(ax[1], math.radians(a), ax[2])
                        vecs.extend([v, v.negative()])
                for v in vecs:
                    if not DraftVecUtils.isNull(v):
                        try:
                            de = Part.LineSegment(last, last.add(v)).toShape()
                        except Part.OCCError:
                            return point, None
                        np = self.getPerpendicular(de, point)
                        if ((self.radius == 0) and (point.sub(last).getAngle(v) < 0.087)) or (
                            (np.sub(point)).Length < self.radius
                        ):
                            if self.tracker and not self.selectMode:
                                self.tracker.setCoords(np)
                                self.tracker.setMarker(self.mk["parallel"])
                                self.tracker.on()
                                self.setCursor("ortho")
                            return np, de
        return point, None

    def snapToGrid(self, point):
        """Return a grid snap point if available."""
        if self.grid:
            if self.grid.Visible:
                if self.isEnabled("Grid"):
                    np = self.grid.getClosestNode(point)
                    if np:
                        dv = point.sub(np)
                        if (self.radius == 0) or (dv.Length <= self.radius):
                            if self.tracker and not self.selectMode:
                                self.tracker.setCoords(np)
                                self.tracker.setMarker(self.mk["grid"])
                                self.tracker.on()
                            self.setCursor("grid")
                            return np
        return point

    def snapToEndpoints(self, shape, point=None):
        """Return a list of endpoints snap locations."""
        if self.isEnabled("Endpoint"):
            if hasattr(shape, "Vertexes"):
                snaps = []
                for v in shape.Vertexes:
                    snaps.append([v.Point, "endpoint", self.toWP(v.Point)])
                return snaps
            if hasattr(shape, "Point"):
                return [[shape.Point, "endpoint", self.toWP(shape.Point)]]
            if hasattr(shape, "Points") and point is not None:
                # point cloud
                # Same as snapToNearUnprojected.
                # Processing individual points in a large point cloud is way too slow:
                # https://github.com/FreeCAD/FreeCAD/issues/22367
                # Must come before handling of mesh as even accessing shape.Points is slow then.
                return [[point, "endpoint", self.toWP(point)]]
            if hasattr(shape, "Points"):
                # mesh
                pts = shape.Points
                if pts and hasattr(pts[0], "Vector"):
                    snaps = []
                    for pt in pts:
                        snaps.append([pt.Vector, "endpoint", self.toWP(pt.Vector)])
                    return snaps
        return []

    def snapToMidpoint(self, shape):
        """Return a list of midpoints snap locations."""
        snaps = []
        if self.isEnabled("Midpoint"):
            if isinstance(shape, Part.Edge):
                mp = DraftGeomUtils.findMidpoint(shape)
                if mp:
                    snaps.append([mp, "midpoint", self.toWP(mp)])
        return snaps

    def snapToNear(self, shape, point):
        """Return a list with a near snap location for an edge."""
        if self.isEnabled("Near") and point:
            try:
                np = shape.Curve.projectPoint(point, "NearestPoint")
            except Exception:
                return []
            return [[np, "passive", self.toWP(np)]]
        else:
            return []

    def snapToNearFace(self, shape, point):
        """Return a list with a near snap location for a face."""
        if self.isEnabled("Near") and point:
            try:
                np = shape.Surface.projectPoint(point, "NearestPoint")
            except Exception:
                return []
            return [[np, "passive", self.toWP(np)]]
        else:
            return []

    def snapToNearUnprojected(self, point):
        """Return a list with a near snap location that is not projected on the object."""
        if self.isEnabled("Near") and point:
            return [[point, "passive", self.toWP(point)]]
        else:
            return []

    def snapToPerpendicular(self, shape, last):
        """Return a list of perpendicular snap locations for an edge."""
        if self.isEnabled("Perpendicular") and last:
            curv = shape.Curve
            try:
                prs = curv.projectPoint(last, "Parameter")
            except Exception:
                return []
            snaps = []
            for pr in prs:
                np = curv.value(pr)
                snaps.append([np, "perpendicular", self.toWP(np)])
            return snaps
        else:
            return []

    def snapToPerpendicularFace(self, shape, last):
        """Return a list of perpendicular snap locations for a face."""
        if self.isEnabled("Perpendicular") and last:
            surf = shape.Surface
            try:
                prs = surf.projectPoint(last, "Parameters")
            except Exception:
                return []
            snaps = []
            for pr in prs:
                np = surf.value(pr[0], pr[1])
                snaps.append([np, "perpendicular", self.toWP(np)])
            return snaps
        else:
            return []

    def snapToOrtho(self, shape, last, constrain):
        """Return a list of ortho snap locations."""
        snaps = []
        if self.isEnabled("Ortho"):
            if constrain:
                if isinstance(shape, Part.Edge):
                    if last:
                        if DraftGeomUtils.geomType(shape) == "Line":
                            if self.constraintAxis:
                                tmpEdge = Part.LineSegment(
                                    last, last.add(self.constraintAxis)
                                ).toShape()
                                # get the intersection points
                                pt = DraftGeomUtils.findIntersection(tmpEdge, shape, True, True)
                                if pt:
                                    for p in pt:
                                        snaps.append([p, "ortho", self.toWP(p)])
        return snaps

    def snapToExtOrtho(self, last, constrain, eline):
        """Return an ortho X extension snap location."""
        if self.isEnabled("Extension") and self.isEnabled("Ortho"):
            if constrain and last and self.constraintAxis and self.extLine:
                tmpEdge1 = Part.LineSegment(last, last.add(self.constraintAxis)).toShape()
                tmpEdge2 = Part.LineSegment(self.extLine.p1(), self.extLine.p2()).toShape()
                # get the intersection points
                pt = DraftGeomUtils.findIntersection(tmpEdge1, tmpEdge2, True, True)
                if pt:
                    return [pt[0], "ortho", pt[0]]
            if eline:
                try:
                    tmpEdge2 = Part.LineSegment(self.extLine.p1(), self.extLine.p2()).toShape()
                    # get the intersection points
                    pt = DraftGeomUtils.findIntersection(eline, tmpEdge2, True, True)
                    if pt:
                        return [pt[0], "ortho", pt[0]]
                except Exception:
                    return None
        return None

    def snapToHold(self, point):
        """Return a snap location that is orthogonal to hold points.

        Or if possible at crossings.
        """
        if not self.holdPoints:
            return None
        wp = self._get_wp()
        u = wp.u
        v = wp.v
        if len(self.holdPoints) > 1:
            # first try mid points
            if self.isEnabled("Midpoint"):
                l = list(self.holdPoints)
                for p1, p2 in itertools.combinations(l, 2):
                    p3 = p1.add((p2.sub(p1)).multiply(0.5))
                    if (p3.sub(point)).Length < self.radius:
                        return [p1, "midpoint", p3]
            # then try int points
            ipoints = []
            l = list(self.holdPoints)
            while len(l) > 1:
                p1 = l.pop()
                for p2 in l:
                    i1 = DraftGeomUtils.findIntersection(p1, p1.add(u), p2, p2.add(v), True, True)
                    if i1:
                        ipoints.append([p1, i1[0]])
                    i2 = DraftGeomUtils.findIntersection(p1, p1.add(v), p2, p2.add(u), True, True)
                    if i2:
                        ipoints.append([p1, i2[0]])
            for p in ipoints:
                if (p[1].sub(point)).Length < self.radius:
                    return [p[0], "ortho", p[1]]
        # then try to stick to a line
        for p in self.holdPoints:
            d = DraftGeomUtils.findDistance(point, [p, p.add(u)])
            if d:
                if d.Length < self.radius:
                    fp = point.add(d)
                    return [p, "extension", fp]
            d = DraftGeomUtils.findDistance(point, [p, p.add(v)])
            if d:
                if d.Length < self.radius:
                    fp = point.add(d)
                    return [p, "extension", fp]
        return None

    def snapToExtPerpendicular(self, last):
        """Return a perpendicular X extension snap location."""
        if self.isEnabled("Extension") and self.isEnabled("Perpendicular"):
            if last and self.extLine:
                if self.extLine.p1() != self.extLine.p2():
                    tmpEdge = Part.LineSegment(self.extLine.p1(), self.extLine.p2()).toShape()
                    np = self.getPerpendicular(tmpEdge, last)
                    return [np, "perpendicular", np]
        return None

    def snapToElines(self, e1, e2):
        """Return a snap at the infinite intersection of the given edges."""
        snaps = []
        if self.isEnabled("Intersection") and self.isEnabled("Extension"):
            if e1 and e2:
                # get the intersection points
                pts = DraftGeomUtils.findIntersection(e1, e2, True, True)
                if pts:
                    for p in pts:
                        snaps.append([p, "intersection", self.toWP(p)])
        return snaps

    def snapToAngles(self, shape):
        """Return a list of angle snap locations."""
        snaps = []
        if self.isEnabled("Angle"):
            place = App.Placement()
            place.Base = shape.Curve.Center
            place.Rotation = App.Rotation(
                App.Vector(1, 0, 0), App.Vector(0, 1, 0), shape.Curve.Axis, "ZXY"
            )
            rad = shape.Curve.Radius
            for deg in (0, 30, 45, 60, 90, 120, 135, 150, 180, 210, 225, 240, 270, 300, 315, 330):
                ang = math.radians(deg)
                cur = App.Vector(math.sin(ang) * rad, math.cos(ang) * rad, 0)
                cur = place.multVec(cur)
                snaps.append([cur, "angle", self.toWP(cur)])
        return snaps

    def snapToCenter(self, shape):
        """Return a list of center snap locations."""
        snaps = []
        if self.isEnabled("Center"):
            cen = shape.Curve.Center
            cen_wp = self.toWP(cen)
            if hasattr(shape.Curve, "Radius"):
                place = App.Placement()
                place.Base = cen
                place.Rotation = App.Rotation(
                    App.Vector(1, 0, 0), App.Vector(0, 1, 0), shape.Curve.Axis, "ZXY"
                )
                rad = shape.Curve.Radius
                for deg in (
                    15,
                    37.5,
                    52.5,
                    75,
                    105,
                    127.5,
                    142.5,
                    165,
                    195,
                    217.5,
                    232.5,
                    255,
                    285,
                    307.5,
                    322.5,
                    345,
                ):
                    ang = math.radians(deg)
                    cur = App.Vector(math.sin(ang) * rad, math.cos(ang) * rad, 0)
                    cur = place.multVec(cur)
                    snaps.append([cur, "center", cen_wp])
            else:
                snaps.append([cen, "center", cen_wp])
        return snaps

    def snapToCenterFace(self, shape):
        """Return a face center snap location."""
        snaps = []
        if self.isEnabled("Center"):
            pos = shape.CenterOfMass
            c = self.toWP(pos)
            snaps.append([pos, "center", c])
        return snaps

    def snapToIntersection(self, shape):
        """Return a list of intersection snap locations."""
        snaps = []
        if self.isEnabled("Intersection"):
            # get the stored objects to calculate intersections
            for obj_name, sub_name in zip(self.lastObj, self.lastObjSubelements):
                obj = App.ActiveDocument.getObject(obj_name)
                if obj and (obj.isDerivedFrom("Part::Feature") or (Draft.getType(obj) == "Axis")):
                    # obj sub is face, shape is edge:
                    if "Face" in sub_name and shape.ShapeType == "Edge":
                        face = obj.Shape.Faces[int(sub_name[4:]) - 1]
                        try:
                            pts = DraftGeomUtils.findIntersection(face, shape)
                            for pt in pts:
                                snaps.append([pt, "intersection", self.toWP(pt)])
                        except Exception:
                            pass
                    elif "Edge" not in sub_name:
                        pass
                    # obj sub is edge, shape is face:
                    elif shape.ShapeType == "Face":
                        edge = obj.Shape.Edges[int(sub_name[4:]) - 1]
                        try:
                            pts = DraftGeomUtils.findIntersection(edge, shape)
                            for pt in pts:
                                snaps.append([pt, "intersection", self.toWP(pt)])
                        except Exception:
                            pass
                    elif shape.ShapeType != "Edge":
                        pass
                    # obj sub is edge, shape is edge:
                    elif (not self.maxEdges) or (len(obj.Shape.Edges) <= self.maxEdges):
                        for edge in obj.Shape.Edges:
                            try:
                                if (
                                    self.isEnabled("WorkingPlane")
                                    and hasattr(edge, "Curve")
                                    and isinstance(edge.Curve, (Part.Line, Part.LineSegment))
                                    and hasattr(shape, "Curve")
                                    and isinstance(shape.Curve, (Part.Line, Part.LineSegment))
                                ):
                                    # get apparent intersection (lines projected on WP)
                                    p1 = self.toWP(edge.Vertexes[0].Point)
                                    p2 = self.toWP(edge.Vertexes[-1].Point)
                                    p3 = self.toWP(shape.Vertexes[0].Point)
                                    p4 = self.toWP(shape.Vertexes[-1].Point)
                                    pts = DraftGeomUtils.findIntersection(
                                        p1, p2, p3, p4, True, True
                                    )
                                else:
                                    pts = DraftGeomUtils.findIntersection(edge, shape)
                                for pt in pts:
                                    snaps.append([pt, "intersection", self.toWP(pt)])
                            except Exception:
                                pass
                                # some curve types yield an error
                                # when trying to read their types
        return snaps

    def snapToPolygon(self, obj):
        """Return a list of polygon center snap locations."""
        snaps = []
        if self.isEnabled("Center"):
            c = obj.Placement.Base
            for edge in obj.Shape.Edges:
                p1 = edge.Vertexes[0].Point
                p2 = edge.Vertexes[-1].Point
                v1 = p1.add((p2 - p1).scale(0.25, 0.25, 0.25))
                v2 = p1.add((p2 - p1).scale(0.75, 0.75, 0.75))
                snaps.append([v1, "center", self.toWP(c)])
                snaps.append([v2, "center", self.toWP(c)])
        return snaps

    def snapToSpecials(self, obj, lastpoint=None, eline=None):
        """Return special snap locations, if any."""
        snaps = []
        if self.isEnabled("Special"):

            if Draft.getType(obj) == "Wall":
                # special snapping for wall: snap to its base shape if it is linear
                if obj.Base:
                    if not obj.Base.Shape.Solids:
                        for v in obj.Base.Shape.Vertexes:
                            snaps.append([v.Point, "special", self.toWP(v.Point)])

            elif Draft.getType(obj) == "Structure":
                # special snapping for struct: only to its base point
                if obj.Base:
                    if not obj.Base.Shape.Solids:
                        for v in obj.Base.Shape.Vertexes:
                            snaps.append([v.Point, "special", self.toWP(v.Point)])
                else:
                    b = obj.Placement.Base
                    snaps.append([b, "special", self.toWP(b)])
                if obj.ViewObject.ShowNodes:
                    for edge in obj.Proxy.getNodeEdges(obj):
                        snaps.extend(self.snapToEndpoints(edge))
                        snaps.extend(self.snapToMidpoint(edge))
                        snaps.extend(self.snapToPerpendicular(edge, lastpoint))
                        snaps.extend(self.snapToIntersection(edge))
                        snaps.extend(self.snapToElines(edge, eline))

            elif hasattr(obj, "SnapPoints"):
                for p in obj.SnapPoints:
                    p2 = obj.Placement.multVec(p)
                    snaps.append([p2, "special", p2])

        return snaps

    def getScreenDist(self, dist, cursor):
        """Return a distance in 3D space from a screen pixels distance."""
        view = Draft.get3DView()
        p1 = view.getPoint(cursor)
        p2 = view.getPoint((cursor[0] + dist, cursor[1]))
        return (p2.sub(p1)).Length

    def getPerpendicular(self, edge, pt):
        """Return a point on an edge, perpendicular to the given point."""
        dv = pt.sub(edge.Vertexes[0].Point)
        nv = DraftVecUtils.project(dv, DraftGeomUtils.vec(edge))
        np = (edge.Vertexes[0].Point).add(nv)
        return np

    def setArchDims(self, p1, p2):
        """Show arc dimensions between 2 points."""
        if self.isEnabled("Dimensions"):
            if not self.dim1:
                self.dim1 = trackers.archDimTracker(mode=2)
            if not self.dim2:
                self.dim2 = trackers.archDimTracker(mode=3)
            self.dim1.p1(p1)
            self.dim2.p1(p1)
            self.dim1.p2(p2)
            self.dim2.p2(p2)
            if self.dim1.Distance:
                self.dim1.on()
            if self.dim2.Distance:
                self.dim2.on()

    def get_quarter_widget(self, mw):
        views = []
        for w in mw.findChild(QtWidgets.QMdiArea).findChildren(QtWidgets.QWidget):
            if w.inherits("SIM::Coin3D::Quarter::QuarterWidget"):
                views.append(w)
        return views

    def device_pixel_ratio(self):
        device_pixel_ratio = 1
        for w in self.get_quarter_widget(Gui.getMainWindow()):
            device_pixel_ratio = w.devicePixelRatio()
        return device_pixel_ratio

    def get_cursor_with_tail(self, base_icon_name, tail_icon_name=None):
        # Other cursor code in scr:
        # src/Gui/CommandView.cpp
        # src/Mod/Mesh/Gui/MeshSelection.cpp
        # src/Mod/Sketcher/Gui/CommandConstraints.cpp

        #   +--------+
        #   |  base  |          vertical offset = 0.5*w
        # w |        +--------+
        #   |   w    |  tail  |
        #   +--------+        | w = width = 16
        #            |   w    |
        #            +--------+

        dpr = self.device_pixel_ratio()
        width = 16 * dpr
        new_icon = QtGui.QPixmap(2 * width, 1.5 * width)
        new_icon.fill(QtCore.Qt.transparent)
        base_icon = QtGui.QPixmap(base_icon_name).scaledToWidth(width)
        qp = QtGui.QPainter()
        qp.begin(new_icon)
        qp.drawPixmap(0, 0, base_icon)
        if tail_icon_name is not None:
            tail_icon = QtGui.QPixmap(tail_icon_name).scaledToWidth(width)
            qp.drawPixmap(width, 0.5 * width, tail_icon)
        qp.end()
        new_icon.setDevicePixelRatio(dpr)
        return QtGui.QCursor(new_icon, 8, 8)

    def setCursor(self, mode=None):
        """Set the cursor to the given mode or unset it."""
        views = self.get_quarter_widget(Gui.getMainWindow())
        if self.selectMode or mode is None:
            self.cursorMode = None
            self.cursorQt = None
            for view in views:
                view.unsetCursor()
        elif self.cursorMode == mode and self.cursorQt is not None:
            for view in views:
                view.setCursor(self.cursorQt)
        else:
            self.cursorMode = mode
            self.cursorQt = self.get_cursor_with_tail(
                ":/icons/Draft_Cursor.svg", None if mode == "passive" else self.cursors[mode]
            )
            for view in views:
                view.setCursor(self.cursorQt)

    def restack(self):
        """Lower the grid tracker so it doesn't obscure other objects."""
        if self.grid:
            self.grid.lowerTracker()

    def off(self):
        """Finish snapping."""
        if self.tracker:
            self.tracker.off()
        if self.trackLine:
            self.trackLine.off()
        if self.extLine:
            self.extLine.off()
        if self.extLine2:
            self.extLine2.off()
        if self.radiusTracker:
            self.radiusTracker.off()
        if self.dim1:
            self.dim1.off()
        if self.dim2:
            self.dim2.off()
        if self.holdTracker:
            self.holdTracker.clear()
            self.holdTracker.off()
        self.unconstrain()
        self.radius = 0
        self.setCursor()
        self.mask = None
        self.selectMode = False
        self.running = False
        self.holdPoints = []
        self.lastObj = []
        self.lastObjSubelements = []

        if hasattr(App, "activeDraftCommand") and App.activeDraftCommand:
            return
        if self.grid:
            if self.grid.show_always is False:
                self.grid.off()
        if params.get_param("SnapBarShowOnlyDuringCommands"):
            toolbar = self.get_snap_toolbar()
            if toolbar:
                toolbar.hide()

    def setSelectMode(self, mode):
        """Set the snapper into select mode (hides snapping temporarily)."""
        self.selectMode = mode
        if not mode:
            self.setCursor()
        else:
            if self.trackLine:
                self.trackLine.off()

    def setAngle(self, delta=None):
        """Keep the current angle."""
        if delta:
            self.mask = delta
        elif isinstance(self.mask, App.Vector):
            self.mask = None
        elif self.trackLine:
            if self.trackLine.Visible:
                self.mask = self.trackLine.p2().sub(self.trackLine.p1())

    def constrain(self, point, basepoint=None, axis=None):
        """Return a constrained point.

        constrain(point,basepoint=None,axis=None: Returns a
        constrained point. Axis can be "x","y" or "z" or a custom vector. If None,
        the closest working plane axis will be picked.
        Basepoint is the base point used to figure out from where the point
        must be constrained. If no basepoint is given, the current point is
        used as basepoint.
        """
        point = App.Vector(point)

        # setup trackers if needed
        if not self.constrainLine:
            self.constrainLine = trackers.lineTracker(dotted=True)

        # setting basepoint
        if not basepoint:
            if not self.basepoint:
                self.basepoint = point
        else:
            self.basepoint = basepoint
        delta = point.sub(self.basepoint)

        if Gui.draftToolBar.globalMode:
            import WorkingPlane

            wp = WorkingPlane.PlaneBase()  # matches the global coordinate system
        else:
            wp = self._get_wp()

        # setting constraint axis
        if axis == "x":
            self.constraintAxis = wp.u
        elif axis == "y":
            self.constraintAxis = wp.v
        elif axis == "z":
            self.constraintAxis = wp.axis
        elif isinstance(axis, App.Vector):
            self.constraintAxis = axis
        else:
            if self.mask is not None:
                self.affinity = self.mask
            if self.affinity is None:
                self.affinity = wp.get_closest_axis(delta)
            if self.affinity == "x":
                self.constraintAxis = wp.u
            elif self.affinity == "y":
                self.constraintAxis = wp.v
            elif self.affinity == "z":
                self.constraintAxis = wp.axis
            elif isinstance(self.affinity, App.Vector):
                self.constraintAxis = self.affinity
            else:
                self.constraintAxis = None

        if self.constraintAxis is None:
            return point

        # calculating constrained point
        cdelta = DraftVecUtils.project(delta, self.constraintAxis)
        npoint = self.basepoint.add(cdelta)

        # setting constrain line
        if self.constrainLine:
            if point != npoint:
                self.constrainLine.p1(point)
                self.constrainLine.p2(npoint)
                self.constrainLine.on()
            else:
                self.constrainLine.off()

        return npoint

    def unconstrain(self):
        """Unset the basepoint and the constrain line."""
        self.basepoint = None
        self.affinity = None
        if self.constrainLine:
            self.constrainLine.off()

    def getPoint(
        self, last=None, callback=None, movecallback=None, extradlg=None, title=None, mode="point"
    ):
        """Get a 3D point from the screen.

        getPoint([last],[callback],[movecallback],[extradlg],[title]):
        gets a 3D point from the screen. You can provide an existing point,
        in that case additional snap options and a tracker are available.
        You can also pass a function as callback, which will get called
        with the resulting point as argument, when a point is clicked,
        and optionally another callback which gets called when
        the mouse is moved.

        If the operation gets cancelled (the user pressed Escape),
        no point is returned.

        Example:

        def cb(point):
            if point:
                print "got a 3D point: ",point

        Gui.Snapper.getPoint(callback=cb)

        If the callback function accepts more than one argument,
        it will also receive the last snapped object. Finally, a qt widget
        can be passed as an extra taskbox.
        title is the title of the point task box mode is the dialog box
        you want (default is point, you can also use wire and line)

        If getPoint() is invoked without any argument, nothing is done
        but the callbacks are removed, so it can be used as a cancel function.
        """
        self.pt = None
        self.holdPoints = []
        self.ui = Gui.draftToolBar
        self.view = Draft.get3DView()

        # remove any previous leftover callbacks
        try:
            if self.callbackClick:
                self.view.removeEventCallbackPivy(
                    coin.SoMouseButtonEvent.getClassTypeId(), self.callbackClick
                )
            if self.callbackMove:
                self.view.removeEventCallbackPivy(
                    coin.SoLocation2Event.getClassTypeId(), self.callbackMove
                )
            if self.callbackClick or self.callbackMove:
                # Next line fixes https://github.com/FreeCAD/FreeCAD/issues/10469:
                gui_utils.end_all_events()
        except RuntimeError:
            # the view has been deleted already
            pass
        self.callbackClick = None
        self.callbackMove = None

        def move(event_cb):
            if not self.ui.mouse:
                return
            event = event_cb.getEvent()
            mousepos = event.getPosition()
            ctrl = event.wasCtrlDown()
            shift = event.wasShiftDown()
            self.pt = Gui.Snapper.snap(mousepos, lastpoint=last, active=ctrl, constrain=shift)
            self.ui.displayPoint(self.pt, last, plane=self._get_wp(), mask=Gui.Snapper.affinity)
            if movecallback:
                movecallback(self.pt, self.snapInfo)

        def getcoords(point, global_mode=True, relative_mode=False):
            """Get the global coordinates from a point."""
            # Same algorithm as in validatePoint in DraftGui.py.
            ref = App.Vector(0, 0, 0)
            if global_mode is False:
                wp = self._get_wp()
                point = wp.get_global_coords(point, as_vector=True)
                ref = wp.get_global_coords(ref)
            if relative_mode is True and last is not None:
                ref = last
            self.pt = point + ref
            accept()

        def click(event_cb):
            if not self.ui.mouse:
                return
            event = event_cb.getEvent()
            if event.getButton() == 1:
                if event.getState() == coin.SoMouseButtonEvent.DOWN:
                    accept()

        def accept():
            try:
                if self.callbackClick:
                    self.view.removeEventCallbackPivy(
                        coin.SoMouseButtonEvent.getClassTypeId(), self.callbackClick
                    )
                if self.callbackMove:
                    self.view.removeEventCallbackPivy(
                        coin.SoLocation2Event.getClassTypeId(), self.callbackMove
                    )
                if self.callbackClick or self.callbackMove:
                    # Next line fixes https://github.com/FreeCAD/FreeCAD/issues/10469:
                    gui_utils.end_all_events()
            except RuntimeError:
                # the view has been deleted already
                pass
            self.callbackClick = None
            self.callbackMove = None
            Gui.Snapper.off()
            self.ui.offUi()
            if callback:
                if len(inspect.getfullargspec(callback).args) > 1:
                    obj = None
                    if self.snapInfo and ("Object" in self.snapInfo) and self.snapInfo["Object"]:
                        obj = App.ActiveDocument.getObject(self.snapInfo["Object"])
                    callback(self.pt, obj)
                else:
                    callback(self.pt)
            self.pt = None

        def cancel():
            try:
                if self.callbackClick:
                    self.view.removeEventCallbackPivy(
                        coin.SoMouseButtonEvent.getClassTypeId(), self.callbackClick
                    )
                if self.callbackMove:
                    self.view.removeEventCallbackPivy(
                        coin.SoLocation2Event.getClassTypeId(), self.callbackMove
                    )
                if self.callbackClick or self.callbackMove:
                    # Next line fixes https://github.com/FreeCAD/FreeCAD/issues/10469:
                    gui_utils.end_all_events()
            except RuntimeError:
                # the view has been deleted already
                pass
            self.callbackClick = None
            self.callbackMove = None
            Gui.Snapper.off()
            self.ui.offUi()
            if callback:
                if len(inspect.getfullargspec(callback).args) > 1:
                    callback(None, None)
                else:
                    callback(None)

        # adding callback functions
        if mode == "line":
            interface = self.ui.lineUi
        elif mode == "wire":
            interface = self.ui.wireUi
        else:
            interface = self.ui.pointUi
        if callback:
            if title:
                interface(
                    title=title, cancel=cancel, getcoords=getcoords, extra=extradlg, rel=bool(last)
                )
            else:
                interface(cancel=cancel, getcoords=getcoords, extra=extradlg, rel=bool(last))
            self.callbackClick = self.view.addEventCallbackPivy(
                coin.SoMouseButtonEvent.getClassTypeId(), click
            )
            self.callbackMove = self.view.addEventCallbackPivy(
                coin.SoLocation2Event.getClassTypeId(), move
            )

    def get_snap_toolbar(self):
        """Get the snap toolbar."""
        if not (hasattr(self, "toolbar") and self.toolbar):
            mw = Gui.getMainWindow()
            self.toolbar = mw.findChild(QtWidgets.QToolBar, "Draft Snap")
        if self.toolbar:
            return self.toolbar

    def toggleGrid(self):
        """Toggle FreeCAD Draft Grid."""
        Gui.runCommand("Draft_ToggleGrid")

    def showradius(self):
        """Show the snap radius indicator."""
        self.radius = self.getScreenDist(params.get_param("snapRange"), (400, 300))
        if self.radiusTracker:
            self.radiusTracker.update(self.radius)
            self.radiusTracker.on()

    def isEnabled(self, snap):
        """Returns true if the given snap is on"""
        if "Lock" in self.active_snaps and snap in self.active_snaps:
            return True
        else:
            return False

    def toggle_snap(self, snap, set_to=None):
        """Sets the given snap on/off according to the given parameter"""
        if set_to:  # set mode
            if set_to is True:
                if not snap in self.active_snaps:
                    self.active_snaps.append(snap)
                status = True
            elif set_to is False:
                if snap in self.active_snaps:
                    self.active_snaps.remove(snap)
                status = False
        else:  # toggle mode, default
            if not snap in self.active_snaps:
                self.active_snaps.append(snap)
                status = True
            elif snap in self.active_snaps:
                self.active_snaps.remove(snap)
                status = False
        self.save_snap_state()
        return status

    def save_snap_state(self):
        """
        Save snap state to user preferences to be restored in next session.
        """
        snap_modes = ""
        for snap in self.snaps:
            if snap in self.active_snaps:
                snap_modes += "1"
            else:
                snap_modes += "0"
        params.set_param("snapModes", snap_modes)

    def show_hide_grids(self, show=True):
        """Show the grid in all 3D views where it was previously visible, or
        hide the grid in all 3D view. Used when switching to different workbenches.

        Hiding the grid can be prevented by setting the GridHideInOtherWorkbenches
        preference to `False`.
        """
        if (not show) and (not params.get_param("GridHideInOtherWorkbenches")):
            return
        mw = Gui.getMainWindow()
        views = mw.getWindowsOfType(
            App.Base.TypeId.fromName("Gui::View3DInventor")
        )  # All 3D views.
        for view in views:
            if view in self.trackers[0]:
                i = self.trackers[0].index(view)
                grid = self.trackers[1][i]
                if show and grid.show_always:
                    grid.on()
                else:
                    grid.off()

    def show(self):
        """Show the grid in all 3D views where it was previously visible."""
        self.show_hide_grids(show=True)

    def hide(self):
        """Hide the grid in all 3D views."""
        self.show_hide_grids(show=False)

    def setGrid(self):
        """Set the grid, if visible."""
        self.setTrackers()

    def setTrackers(self, update_grid=True):
        """Set the trackers."""
        v = Draft.get3DView()
        if v is None:
            return

        if v != self.activeview:
            if v in self.trackers[0]:
                i = self.trackers[0].index(v)
                self.grid = self.trackers[1][i]
                self.tracker = self.trackers[2][i]
                self.extLine = self.trackers[3][i]
                self.radiusTracker = self.trackers[4][i]
                self.dim1 = self.trackers[5][i]
                self.dim2 = self.trackers[6][i]
                self.trackLine = self.trackers[7][i]
                self.extLine2 = self.trackers[8][i]
                self.holdTracker = self.trackers[9][i]
            else:
                self.grid = trackers.gridTracker()
                if params.get_param("alwaysShowGrid"):
                    self.grid.show_always = True
                if params.get_param("grid"):
                    self.grid.show_during_command = True
                self.tracker = trackers.snapTracker()
                self.trackLine = trackers.lineTracker()
                self.extLine = trackers.lineTracker(dotted=True)
                self.extLine2 = trackers.lineTracker(dotted=True)
                self.radiusTracker = trackers.radiusTracker()
                self.dim1 = trackers.archDimTracker(mode=2)
                self.dim2 = trackers.archDimTracker(mode=3)
                self.holdTracker = trackers.snapTracker()
                self.holdTracker.setMarker("cross")
                self.holdTracker.clear()
                self.trackers[0].append(v)
                self.trackers[1].append(self.grid)
                self.trackers[2].append(self.tracker)
                self.trackers[3].append(self.extLine)
                self.trackers[4].append(self.radiusTracker)
                self.trackers[5].append(self.dim1)
                self.trackers[6].append(self.dim2)
                self.trackers[7].append(self.trackLine)
                self.trackers[8].append(self.extLine2)
                self.trackers[9].append(self.holdTracker)
            self.activeview = v

        if not update_grid:
            return

        if self.grid.show_always or (
            self.grid.show_during_command
            and hasattr(App, "activeDraftCommand")
            and App.activeDraftCommand
        ):
            self.grid.set()

    def addHoldPoint(self):
        """Add hold snap point to list of hold points."""
        if self.spoint and self.spoint not in self.holdPoints:
            if self.holdTracker:
                self.holdTracker.addCoords(self.spoint)
                self.holdTracker.setColor()
                self.holdTracker.on()
            self.holdPoints.append(self.spoint)

    def recenter_workingplane(self):
        """Recenters the working plane on the current snap position"""
        if self.spoint:
            self._get_wp().set_to_position(self.toWP(self.spoint))


## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2017 sliptonic <shopinthewoods@gmail.com>               *
# *   Copyright (c) 2020 russ4262 (Russell Johnson)                         *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Path
import Path.Op.Area as PathAreaOp
import Path.Op.Base as PathOp

from PySide.QtCore import QT_TRANSLATE_NOOP


__title__ = "Base CAM Pocket Operation"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Base class and implementation for pocket operations."

if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())

translate = FreeCAD.Qt.translate


class ObjectPocket(PathAreaOp.ObjectOp):
    """Base class for proxy objects of all pocket operations."""

    @classmethod
    def pocketPropertyEnumerations(cls, dataType="data"):
        """pocketPropertyEnumerations(dataType="data")... return property enumeration lists of specified dataType.
        Args:
            dataType = 'data', 'raw', 'translated'
        Notes:
        'data' is list of internal string literals used in code
        'raw' is list of (translated_text, data_string) tuples
        'translated' is list of translated string literals
        """

        enums = {
            "CutMode": [
                (translate("CAM_Pocket", "Climb"), "Climb"),
                (translate("CAM_Pocket", "Conventional"), "Conventional"),
            ],  # this is the direction that the profile runs
            "StartAt": [
                (translate("CAM_Pocket", "Center"), "Center"),
                (translate("CAM_Pocket", "Edge"), "Edge"),
            ],
            "ClearingPattern": [
                (translate("CAM_Pocket", "ZigZag"), "ZigZag"),
                (translate("CAM_Pocket", "Offset"), "Offset"),
                (translate("CAM_Pocket", "ZigZagOffset"), "ZigZagOffset"),
                (translate("CAM_Pocket", "Line"), "Line"),
                (translate("CAM_Pocket", "Grid"), "Grid"),
            ],  # Fill Pattern
        }

        if dataType == "raw":
            return enums

        data = list()
        idx = 0 if dataType == "translated" else 1

        Path.Log.debug(enums)

        for k, v in enumerate(enums):
            data.append((v, [tup[idx] for tup in enums[v]]))
        Path.Log.debug(data)

        return data

    def areaOpFeatures(self, obj):
        """areaOpFeatures(obj) ... Pockets have a FinishDepth and work on Faces"""
        return PathOp.FeatureBaseFaces | PathOp.FeatureFinishDepth | self.pocketOpFeatures(obj)

    def pocketOpFeatures(self, obj):
        return 0

    def initPocketOp(self, obj):
        """initPocketOp(obj) ... overwrite to initialize subclass.
        Can safely be overwritten by subclass."""
        pass

    def opExecute(self, obj):
        if len(obj.Base) == 0:
            return
        super().opExecute(obj)

    def areaOpSetDefaultValues(self, obj, job):
        obj.PocketLastStepOver = 0

    def pocketInvertExtraOffset(self):
        """pocketInvertExtraOffset() ... return True if ExtraOffset's direction is inward.
        Can safely be overwritten by subclass."""
        return False

    def initAreaOp(self, obj):
        """initAreaOp(obj) ... create pocket specific properties.
        Do not overwrite, implement initPocketOp(obj) instead."""
        Path.Log.track()

        # Pocket Properties
        obj.addProperty(
            "App::PropertyEnumeration",
            "CutMode",
            "Pocket",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "The direction that the toolpath should go around the part ClockWise (CW) or CounterClockWise (CCW)",
            ),
        )
        obj.addProperty(
            "App::PropertyDistance",
            "ExtraOffset",
            "Pocket",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Extra offset to apply to the operation. Direction is operation dependent.",
            ),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "StartAt",
            "Pocket",
            QT_TRANSLATE_NOOP("App::Property", "Start pocketing at center or boundary"),
        )
        obj.addProperty(
            "App::PropertyPercent",
            "StepOver",
            "Pocket",
            QT_TRANSLATE_NOOP(
                "App::Property", "Percent of cutter diameter to step over on each pass"
            ),
        )
        obj.addProperty(
            "App::PropertyFloat",
            "ZigZagAngle",
            "Pocket",
            QT_TRANSLATE_NOOP("App::Property", "Angle of the zigzag pattern"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "ClearingPattern",
            "Pocket",
            QT_TRANSLATE_NOOP("App::Property", "Clearing pattern to use"),
        )
        obj.addProperty(
            "App::PropertyBool",
            "MinTravel",
            "Pocket",
            QT_TRANSLATE_NOOP("App::Property", "Use 3D Sorting of Path"),
        )
        obj.addProperty(
            "App::PropertyBool",
            "KeepToolDown",
            "Pocket",
            QT_TRANSLATE_NOOP("App::Property", "Attempts to avoid unnecessary retractions."),
        )
        obj.addProperty(
            "App::PropertyPercent",
            "PocketLastStepOver",
            "Pocket",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Last Stepover Radius.  If 0, 50% of cutter is used. Tuning this can be used to improve stepover for some shapes",
            ),
        )
        obj.addProperty(
            "App::PropertyBool",
            "UseRestMachining",
            "Pocket",
            QT_TRANSLATE_NOOP(
                "App::Property",
                "Skips machining regions that have already been cleared by previous operations.",
            ),
        )

        for n in self.pocketPropertyEnumerations():
            setattr(obj, n[0], n[1])

        self.initPocketOp(obj)

    def areaOpRetractTool(self, obj):
        Path.Log.debug("retracting tool: %d" % (not obj.KeepToolDown))
        return not obj.KeepToolDown

    def areaOpUseProjection(self, obj):
        """areaOpUseProjection(obj) ... return False"""
        return False

    def areaOpAreaParams(self, obj, isHole):
        """areaOpAreaParams(obj, isHole) ... return dictionary with pocket's area parameters"""
        Path.Log.track()
        params = {}
        params["Fill"] = 0
        params["Coplanar"] = 0
        params["PocketMode"] = 1
        params["SectionCount"] = -1
        params["Angle"] = obj.ZigZagAngle
        params["FromCenter"] = obj.StartAt == "Center"
        params["PocketStepover"] = (self.radius * 2) * (float(obj.StepOver) / 100)
        extraOffset = obj.ExtraOffset.Value
        if self.pocketInvertExtraOffset():
            extraOffset = 0 - extraOffset
        params["PocketExtraOffset"] = extraOffset
        params["ToolRadius"] = self.radius
        params["PocketLastStepover"] = obj.PocketLastStepOver

        Pattern = {
            "ZigZag": 1,
            "Offset": 2,
            "ZigZagOffset": 4,
            "Line": 5,
            "Grid": 6,
        }

        params["PocketMode"] = Pattern.get(obj.ClearingPattern, 1)

        if obj.SplitArcs:
            params["Explode"] = True
            params["FitArcs"] = False

        return params

    def opOnDocumentRestored(self, obj):
        super().opOnDocumentRestored(obj)
        if not hasattr(obj, "PocketLastStepOver"):
            obj.addProperty(
                "App::PropertyPercent",
                "PocketLastStepOver",
                "Pocket",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Last Stepover Radius.  If 0, 50% of cutter is used. Tuning this can be used to improve stepover for some shapes",
                ),
            )
            obj.PocketLastStepOver = 0

        if not hasattr(obj, "UseRestMachining"):
            obj.addProperty(
                "App::PropertyBool",
                "UseRestMachining",
                "Pocket",
                QT_TRANSLATE_NOOP(
                    "App::Property",
                    "Skips machining regions that have already been cleared by previous operations.",
                ),
            )

        if hasattr(obj, "OffsetPattern"):
            obj.setGroupOfProperty("OffsetPattern", "Pocket")
            obj.renameProperty("OffsetPattern", "ClearingPattern")
        if hasattr(obj, "RestMachiningRegions"):
            obj.removeProperty("RestMachiningRegions")
        if hasattr(obj, "RestMachiningRegionsNeedRecompute"):
            obj.removeProperty("RestMachiningRegionsNeedRecompute")

        Path.Log.track()

    def areaOpPathParams(self, obj, isHole):
        """areaOpAreaParams(obj, isHole) ... return dictionary with pocket's path parameters"""
        params = {}

        CutMode = ["Conventional", "Climb"]
        params["orientation"] = CutMode.index(obj.CutMode)

        # if MinTravel is turned on, set path sorting to 3DSort
        # 3DSort shouldn't be used without a valid start point. Can cause
        # tool crash without it.
        #
        # ml: experimental feature, turning off for now (see https://forum.freecad.org/viewtopic.php?f=15&t=24422&start=30#p192458)
        # realthunder: I've fixed it with a new sorting algorithm, which I
        # tested fine, but of course need more test. Please let know if there is
        # any problem
        #
        if obj.MinTravel and obj.UseStartPoint and obj.StartPoint is not None:
            params["sort_mode"] = 3
            params["threshold"] = self.radius * 2
        return params


def SetupProperties():
    setup = PathAreaOp.SetupProperties()
    setup.append("CutMode")
    setup.append("ExtraOffset")
    setup.append("StepOver")
    setup.append("ZigZagAngle")
    setup.append("ClearingPattern")
    setup.append("StartAt")
    setup.append("MinTravel")
    setup.append("KeepToolDown")
    return setup



# SPDX-License-Identifier: LGPL-2.1-or-later

from FreeCAD import Base
import importlib

def OpenInsertObject(importerName, objectPath, importMethod, docName = ""):
    try:
        importArgs = []
        importKwargs = {}

        importerModule = importlib.import_module(importerName)

        if docName:
            importArgs.append(docName)
        if hasattr(importerModule, "importOptions"):
            importKwargs["options"] = importerModule.importOptions(objectPath)

        getattr(importerModule, importMethod)(objectPath, *importArgs, **importKwargs)
    except Base.AbortIOException:
        pass
